=== modified file 'doc/setup/install_proc.txt'
--- doc/setup/install_proc.txt	2009-01-14 22:42:10 +0000
+++ doc/setup/install_proc.txt	2009-02-04 12:51:17 +0000
@@ -21,9 +21,9 @@
 ##########################################################################
 sudo apt-get install rsync subversion build-essential debootstrap       \
                      apache2 libapache2-mod-python libapache2-svn       \
-                     python2.5-dev python-svn python-webpy python-cjson \
-                     postgresql python-pygresql php5 php5-pgsql         \
-                     python-docutils python-epydoc python-ldap          \
+                     python-svn python-cjson python-genshi              \
+                     postgresql php5 php5-pgsql python-routes           \
+                     python-docutils python-epydoc                      \
                      python-storm python-psycopg2
 
 ##########################################################################

=== modified file 'ivle/dispatch/__init__.py'
--- ivle/dispatch/__init__.py	2009-01-22 01:21:48 +0000
+++ ivle/dispatch/__init__.py	2009-02-04 12:54:12 +0000
@@ -1,5 +1,5 @@
-# IVLE
-# Copyright (C) 2007-2008 The University of Melbourne
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -15,16 +15,16 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-# Module: dispatch
-# Author: Matt Giuca
-# Date: 11/12/2007
-
-# This is a mod_python handler program. The correct way to call it is to have
-# Apache send all requests to be handled by the module 'dispatch'.
-
-# Top-level handler. Handles all requests to all pages in IVLE.
-# Handles authentication (not authorization).
-# Then passes the request along to the appropriate ivle app.
+# Author: Matt Giuca, Will Grant
+
+"""
+This is a mod_python handler program. The correct way to call it is to have
+Apache send all requests to be handled by the module 'dispatch'.
+
+Top-level handler. Handles all requests to all pages in IVLE.
+Handles authentication (not authorization).
+Then passes the request along to the appropriate ivle app.
+"""
 
 import sys
 import os
@@ -37,17 +37,46 @@
 import time
 
 import mod_python
-from mod_python import apache, Cookie
+import routes
 
 from ivle import util
 import ivle.conf
 import ivle.conf.apps
+from ivle.dispatch.request import Request
+from ivle.dispatch import login
 import apps
-import login
 import html
-from request import Request
 import plugins.console # XXX: Relies on www/ being in the Python path.
 
+# XXX List of plugins, which will eventually be read in from conf
+plugins_HACK = [
+    'ivle.webapp.admin.user#Plugin',
+    'ivle.webapp.tutorial#Plugin',
+    'ivle.webapp.admin.subject#Plugin',
+    'ivle.webapp.browser#Plugin',
+    'ivle.webapp.groups#Plugin',
+]
+
+def get_routes_mapper():
+    """
+    Build a Mapper object for doing URL matching using 'routes', based on the
+    plugins config.
+    """
+    m = routes.Mapper(explicit=True)
+    for pluginstr in plugins_HACK:
+        plugin_path, classname = pluginstr.split('#')
+        # Load the plugin module from somewhere in the Python path
+        # (Note that plugin_path is a fully-qualified Python module name).
+        plugin = getattr(__import__(plugin_path, fromlist=[classname]),
+            classname)
+        # Establish a URL pattern for each element of plugin.urls
+        for url in plugin.urls:
+            routex = url[0]
+            view_class = url[1]
+            kwargs_dict = url[2] if len(url) >= 3 else {}
+            m.connect(routex, view=view_class, **kwargs_dict)
+    return m
+
 def handler(req):
     """Handles a request which may be to anywhere in the site except media.
     Intended to be called by mod_python, as a handler.
@@ -63,7 +92,7 @@
         # yet.
         handle_unknown_exception(apachereq, *sys.exc_info())
         # Tell Apache not to generate its own errors as well
-        return apache.OK
+        return mod_python.apache.OK
 
     # Run the main handler, and catch all exceptions
     try:
@@ -75,7 +104,7 @@
     except Exception:
         handle_unknown_exception(req, *sys.exc_info())
         # Tell Apache not to generate its own errors as well
-        return apache.OK
+        return mod_python.apache.OK
 
 def handler_(req, apachereq):
     """
@@ -88,6 +117,26 @@
     if not req.publicmode:
         req.user = login.get_user_details(req)
 
+    ### BEGIN New plugins framework ###
+    # XXX This should be done ONCE per Python process, not per request.
+    # (Wait till WSGI)
+    # XXX No authentication is done here
+    req.mapper = get_routes_mapper()
+    matchdict = req.mapper.match(req.uri)
+    if matchdict is not None:
+        viewcls = matchdict['view']
+        # Get the remaining arguments, less 'view', 'action' and 'controller'
+        # (The latter two seem to be built-in, and we don't want them).
+        kwargs = matchdict.copy()
+        del kwargs['view']
+        # Instantiate the view, which should be a BaseView class
+        view = viewcls(req, **kwargs)
+        # Render the output
+        view.render(req)
+        req.store.commit()
+        return req.OK
+    ### END New plugins framework ###
+
     # Check req.app to see if it is valid. 404 if not.
     if req.app is not None and req.app not in ivle.conf.apps.app_url:
         req.throw_error(Request.HTTP_NOT_FOUND,
@@ -180,10 +229,10 @@
     # For some reason, some versions of mod_python have "_server" instead of
     # "main_server". So we check for both.
     try:
-        admin_email = apache.main_server.server_admin
+        admin_email = mod_python.apache.main_server.server_admin
     except AttributeError:
         try:
-            admin_email = apache._server.server_admin
+            admin_email = mod_python.apache._server.server_admin
         except AttributeError:
             admin_email = ""
     try:
@@ -191,7 +240,7 @@
         req.status = httpcode
     except AttributeError:
         httpcode = None
-        req.status = apache.HTTP_INTERNAL_SERVER_ERROR
+        req.status = mod_python.apache.HTTP_INTERNAL_SERVER_ERROR
     try:
         publicmode = req.publicmode
     except AttributeError:
@@ -301,8 +350,8 @@
         logging.error('%s\n%s'%(str(msg), tb))
         # Error messages are only displayed is the user is NOT a student,
         # or if there has been a problem logging the error message
-        show_errors = (not publicmode) and (not login or \
-                            (not str(role) == "student") or logfail)
+        show_errors = (not publicmode) and ((login and \
+                            str(role) != "student") or logfail)
         req.write("""<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"                 
 	"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">                                      
 <html xmlns="http://www.w3.org/1999/xhtml">
@@ -311,7 +360,7 @@
 <h1>IVLE Internal Server Error""")
         if (show_errors):
             if (codename is not None
-                        and httpcode != apache.HTTP_INTERNAL_SERVER_ERROR):
+                        and httpcode != mod_python.apache.HTTP_INTERNAL_SERVER_ERROR):
                 req.write(": %s" % cgi.escape(codename))
         
         req.write("""</h1>

=== modified file 'ivle/dispatch/login.py'
--- ivle/dispatch/login.py	2009-01-15 01:18:00 +0000
+++ ivle/dispatch/login.py	2009-02-03 23:49:34 +0000
@@ -23,8 +23,6 @@
 import os
 import datetime
 
-from mod_python import Session
-
 import ivle.conf
 from ivle import (util, caps, forumutil)
 from ivle.auth import authenticate, AuthError

=== modified file 'ivle/dispatch/request.py'
--- ivle/dispatch/request.py	2009-01-19 08:14:25 +0000
+++ ivle/dispatch/request.py	2009-02-03 23:49:34 +0000
@@ -25,8 +25,13 @@
 object.
 """
 
-import mod_python
-from mod_python import (util, Session, Cookie)
+try:
+    import mod_python.Session
+    import mod_python.Cookie
+    import mod_python.util
+except ImportError:
+    # This needs to be importable from outside Apache.
+    pass
 
 import ivle.util
 import ivle.conf
@@ -277,10 +282,10 @@
             self.session.invalidate()
             self.session.delete()
             # Invalidates all IVLE cookies
-            all_cookies = Cookie.get_cookies(self)
+            all_cookies = mod_python.Cookie.get_cookies(self)
             for cookie in all_cookies:
                 if cookie in ivle_cookies:
-                    self.add_cookie(Cookie.Cookie(cookie,'',expires=1,path='/'))
+                    self.add_cookie(mod_python.Cookie.Cookie(cookie,'',expires=1,path='/'))
         self.throw_redirect(ivle.util.make_path('')) 
 
 
@@ -327,9 +332,9 @@
     def add_cookie(self, cookie, value=None, **attributes):
         """Inserts a cookie into this request object's headers."""
         if value is None:
-            Cookie.add_cookie(self.apache_req, cookie)
+            mod_python.Cookie.add_cookie(self.apache_req, cookie)
         else:
-            Cookie.add_cookie(self.apache_req, cookie, value, **attributes)
+            mod_python.Cookie.add_cookie(self.apache_req, cookie, value, **attributes)
 
     def get_session(self):
         """Returns a mod_python Session object for this request.
@@ -337,7 +342,7 @@
         interface if porting away from mod_python."""
         # Cache the session object and set the timeout to 24 hours.
         if not hasattr(self, 'session'):
-            self.session = Session.FileSession(self.apache_req,
+            self.session = mod_python.Session.FileSession(self.apache_req,
                                                timeout = 60 * 60 * 24)
         return self.session
 
@@ -347,7 +352,7 @@
         interface if porting away from mod_python."""
         # Cache the fieldstorage object
         if not hasattr(self, 'fields'):
-            self.fields = util.FieldStorage(self.apache_req)
+            self.fields = mod_python.util.FieldStorage(self.apache_req)
         return self.fields
 
     def get_cgi_environ(self):

=== modified file 'ivle/makeuser.py'
--- ivle/makeuser.py	2009-02-02 05:14:04 +0000
+++ ivle/makeuser.py	2009-01-20 01:02:29 +0000
@@ -240,62 +240,47 @@
         # Chmod to rwxr-xr-x (755)
         os.chmod(userhomedir, 0755)
 
-    make_conf_py(user.login, userdir, user.svn_pass)
+    make_conf_py(user.login, userdir, ivle.conf.jail_system, user.svn_pass)
     make_etc_passwd(user.login, userdir, ivle.conf.jail_system, user.unixid)
 
     return userhomedir
 
-def make_conf_py(username, user_jail_dir, svn_pass):
+def make_conf_py(username, user_jail_dir, staging_dir, svn_pass):
     """
     Creates (overwriting any existing file, and creating directories) a
     file ${python_site_packages}/ivle/conf/conf.py in a given user's jail.
     username: Username.
     user_jail_dir: User's jail dir, ie. ivle.conf.jail_base + username
-    svn_pass: User's SVN password.
+    staging_dir: The dir with the staging copy of the jail. (With the
+        template conf.py file).
+    svn_pass: As with make_jail. User's SVN password, but if not supplied,
+        will look up in the DB.
     """
+    template_conf_path = os.path.join(staging_dir,
+            ivle.conf.python_site_packages[1:], "ivle/conf/conf.py")
     conf_path = os.path.join(user_jail_dir,
             ivle.conf.python_site_packages[1:], "ivle/conf/conf.py")
     os.makedirs(os.path.dirname(conf_path))
 
-    # In the "in-jail" version of conf, we don't need MOST of the details
-    # (it would be a security risk to have them here).
-    # So we just write root_dir, and jail_base is "/".
-    # (jail_base being "/" means "jail-relative" paths are relative to "/"
-    # when inside the jail.)
+    # Read the contents of the template conf file
+    try:
+        template_conf_file = open(template_conf_path, "r")
+        template_conf_data = template_conf_file.read()
+        template_conf_file.close()
+    except:
+        # Couldn't open template conf.py for some reason
+        # Just treat it as empty file
+        template_conf_data = ("# Warning: Problem building config script.\n"
+                              "# Could not find template conf.py file.\n")
 
-    # XXX: jail_base is wrong and shouldn't be here. Unfortunately, jail code
-    #      uses ivle.studpath.url_to_{local,jailpaths}, both of which use
-    #      jail_base. Note that they don't use the bits of the return value
-    #      that depend on jail_base, so it can be any string.
     conf_file = open(conf_path, "w")
-    conf_file.write("""# IVLE jail configuration
-
-# In URL space, where in the site is IVLE located. (All URLs will be prefixed
-# with this).
-# eg. "/" or "/ivle".
-root_dir = %(root_dir)r
-
-# This value is not relevant inside the jail, but must remain for now. See
-# the XXX in ivle.makeuser.make_conf_py.
-jail_base = '/'
-
-# The hostname for serving publicly accessible pages
-public_host = %(public_host)r
-
-# The URL under which the Subversion repositories are located.
-svn_addr = %(svn_addr)r
-
-# The login name for the owner of the jail
-login = %(username)r
-
-# The subversion-only password for the owner of the jail
-svn_pass = %(svn_pass)r
-""" % {'root_dir': ivle.conf.root_dir,
-       'public_host': ivle.conf.public_host,
-       'svn_addr': ivle.conf.svn_addr,
-       'username': username,
-       'svn_pass': svn_pass,
-      })
+    conf_file.write(template_conf_data)
+    conf_file.write("\n# The login name for the owner of the jail\n")
+    conf_file.write("login = %s\n" % repr(username))
+    conf_file.write("\n")
+    conf_file.write("# The subversion-only password for the owner of "
+        "the jail\n")
+    conf_file.write("svn_pass = %s\n" % repr(svn_pass))
     conf_file.close()
 
     # Make this file world-readable

=== modified file 'ivle/studpath.py'
--- ivle/studpath.py	2009-01-13 01:36:15 +0000
+++ ivle/studpath.py	2009-02-04 01:54:46 +0000
@@ -81,7 +81,7 @@
     urlpath: See urlpath in url_to_local.
 
     >>> url_to_jailpaths("joe/mydir/myfile")
-    ("joe", "/home/informatics/jails/joe", "home/joe/mydir/myfile")
+    ('joe', '/var/lib/ivle/jailmounts/joe', '/home/joe/mydir/myfile')
 
     >>> url_to_jailpaths("")
     (None, None, None)

=== added directory 'ivle/webapp'
=== added file 'ivle/webapp/__init__.py'
=== added directory 'ivle/webapp/admin'
=== added file 'ivle/webapp/admin/__init__.py'
=== added file 'ivle/webapp/admin/subject.html'
--- ivle/webapp/admin/subject.html	1970-01-01 00:00:00 +0000
+++ ivle/webapp/admin/subject.html	2009-02-04 07:18:24 +0000
@@ -0,0 +1,8 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <head>
+  </head>
+  <body>
+    <object class="fullscreen" type="text/html" data="${serve_loc}"></object>
+  </body>
+</html>

=== renamed file 'www/apps/subjects/__init__.py' => 'ivle/webapp/admin/subject.py'
--- www/apps/subjects/__init__.py	2009-01-15 06:11:02 +0000
+++ ivle/webapp/admin/subject.py	2009-02-04 12:54:12 +0000
@@ -26,60 +26,65 @@
 import urllib
 import cgi
 
-import ivle.database
+from ivle.webapp.base.views import XHTMLView
+from ivle.webapp.base.plugins import BasePlugin
+from ivle.webapp.errors import NotFound
+from ivle.database import Subject
 from ivle import util
 
-def handle(req):
-    """Handler for the Subjects application. Links to subject home pages."""
-
-    req.styles = ["media/subjects/subjects.css"]
-    if req.path == "":
-        handle_toplevel_menu(req)
-    else:
-        handle_subject_page(req, req.path)
-
-def handle_toplevel_menu(req):
-    # This is represented as a directory. Redirect and add a slash if it is
-    # missing.
-    if req.uri[-1] != '/':
-        req.throw_redirect(req.uri + '/')
-
-    enrolled_subjects = req.user.subjects
-    unenrolled_subjects = [subject for subject in
-                           req.store.find(ivle.database.Subject)
-                           if subject not in enrolled_subjects]
-
-    def print_subject(subject):
-        if subject.url is None:
-            req.write('  <li>%s (no home page)</li>\n'
-                % cgi.escape(subject.name))
-        else:
-            req.write('  <li><a href="%s">%s</a></li>\n'
-                % (cgi.escape(subject.url),
-                   cgi.escape(subject.name)))
-
-    req.content_type = "text/html"
-    req.write_html_head_foot = True
-    req.write('<div id="ivle_padding">\n')
-    req.write("<h2>IVLE Subject Homepages</h2>\n")
-    req.write("<h2>Subjects</h2>\n<ul>\n")
-    for subject in enrolled_subjects:
-        print_subject(subject)
-    req.write("</ul>\n")
-    if len(unenrolled_subjects) > 0:
-        req.write("<h3>Other Subjects</h3>\n")
-        req.write("<p>You are not currently enrolled in these subjects</p>\n")
-        req.write("<ul>\n")
-        for subject in unenrolled_subjects:
-            print_subject(subject)
-        req.write("</ul>\n")
-    req.write("</div>\n")
-
-def handle_subject_page(req, path):
-    req.content_type = "text/html"
-    req.write_html_head_foot = True     # Have dispatch print head and foot
-
-    # Just make the iframe pointing to media/subjects
-    serve_loc = util.make_path(os.path.join('media', 'subjects', path))
-    req.write('<object class="fullscreen" type="text/html" \
-data="%s"></iframe>'% urllib.quote(serve_loc))
+
+class SubjectsView(XHTMLView):
+    '''The view of the list of subjects.'''
+    app_template = 'subjects.html'
+    appname = 'subjects'
+
+    def populate(self, req, ctx):
+        req.styles = ["media/subjects/subjects.css"]
+
+        enrolled_subjects = req.user.subjects
+        unenrolled_subjects = [subject for subject in
+                               req.store.find(Subject)
+                               if subject not in enrolled_subjects]
+
+        ctx['enrolled_subjects'] = []
+        ctx['other_subjects'] = []
+
+        req.content_type = "text/html"
+        req.write_html_head_foot = True
+
+        for subject in enrolled_subjects:
+            new_subj = {}
+            new_subj['name'] = subject.name
+            new_subj['url'] = subject.url
+            ctx['enrolled_subjects'].append(new_subj)
+
+        if len(unenrolled_subjects) > 0:
+            for subject in unenrolled_subjects:
+                new_subj = {}
+                new_subj['name'] = subject.name
+                new_subj['url'] = subject.url
+                ctx['other_subjects'].append(new_subj)
+
+
+class SubjectView(XHTMLView):
+    '''The view of a subject.'''
+    app_template = 'subject.html'
+    appname = 'subjects' # XXX
+
+    def __init__(self, req, subject, path):
+        self.subject = req.store.find(Subject, code=subject).one()
+        self.path = path
+
+    def populate(self, req, ctx):
+        if self.subject is None:
+            raise NotFound()
+
+        ctx['serve_loc'] = urllib.quote(util.make_path(os.path.join('media',
+                                    'subjects', self.subject.code, self.path)))
+
+class Plugin(BasePlugin):
+    urls = [
+        ('subjects/', SubjectsView),
+        ('subjects/:subject', SubjectView, {'path': ''}),
+        ('subjects/:subject/*(path)', SubjectView),
+    ]

=== added file 'ivle/webapp/admin/subjects.html'
--- ivle/webapp/admin/subjects.html	1970-01-01 00:00:00 +0000
+++ ivle/webapp/admin/subjects.html	2009-02-04 07:18:24 +0000
@@ -0,0 +1,42 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <head>
+    <title>Subjects</title>
+  </head>
+  <body>
+    <div id="ivle_padding">
+      <h2>IVLE Subject Homepages</h2>
+      <py:if test="(len(enrolled_subjects) > 0)">
+        <h2>Subjects</h2>
+        <ul>
+          <py:for each="subject in enrolled_subjects">
+            <py:choose test="subject['url']">
+              <py:when test="None">
+                <li>${subject['name']} (no home page)</li>
+              </py:when>
+              <py:otherwise>
+                <li><a href="${subject['url']}">${subject['name']}</a></li>
+              </py:otherwise>
+            </py:choose>
+          </py:for>
+        </ul>
+      </py:if>
+      <py:if test="(len(other_subjects) > 0)">
+        <h3>Other Subjects</h3>
+        <p>You are not currently enrolled in these subjects</p>
+        <ul>
+          <py:for each="subject in other_subjects">
+            <py:choose test="subject['url']">
+             <py:when test="None">
+               <li>${subject['name']} (no home page)</li>
+             </py:when>
+             <py:otherwise>
+               <li><a href="${subject['url']}">${subject['name']}</a></li>
+             </py:otherwise>
+            </py:choose>
+          </py:for>
+        </ul>
+      </py:if>
+    </div>
+  </body>
+</html>

=== added file 'ivle/webapp/admin/user.py'
--- ivle/webapp/admin/user.py	1970-01-01 00:00:00 +0000
+++ ivle/webapp/admin/user.py	2009-02-03 22:46:32 +0000
@@ -0,0 +1,76 @@
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+
+# Author: Matt Giuca, Will Grant
+
+from ivle.webapp.base.views import JSONRESTView
+from ivle.webapp.base.plugins import BasePlugin
+import ivle.database
+import ivle.util
+
+# List of fields returned as part of the user JSON dictionary
+# (as returned by the get_user action)
+user_fields_list = (
+    "login", "state", "unixid", "email", "nick", "fullname",
+    "rolenm", "studentid", "acct_exp", "pass_exp", "last_login",
+    "svn_pass"
+)
+
+class UserRESTView(JSONRESTView):
+    """
+    A REST interface to the user object.
+    """
+    def __init__(self, req, login):
+        super(UserRESTView, self).__init__(self, req, login)
+        self.context = ivle.database.User.get_by_login(req.store, login)
+
+    def GET(self, req):
+        # XXX Check Caps
+        user = ivle.util.object_to_dict(user_fields_list, self.context)
+        # Convert time stamps to nice strings
+        for k in 'pass_exp', 'acct_exp', 'last_login':
+            if user[k] is not None:
+                user[k] = unicode(user[k])
+
+        user['local_password'] = self.context.passhash is not None
+        return user
+
+    def PATCH(self, req, data):
+        # XXX Check Caps
+        # XXX Admins can set extra fields
+        # Note: Cannot change password here (use change_password named op)
+
+        for f in user_fields_list:
+            try:
+                field = data[f]
+                if isinstance(field, str):
+                    field = unicode(field)
+                setattr(self.context, f, field)
+            except KeyError:
+                continue
+
+class Plugin(BasePlugin):
+    """
+    The Plugin class for the user plugin.
+    """
+    # Magic attribute: urls
+    # Sequence of pairs/triples of
+    # (regex str, handler class, kwargs dict)
+    # The kwargs dict is passed to the __init__ of the view object
+    urls = [
+        ('api/users/:login', UserRESTView)
+    ]

=== added directory 'ivle/webapp/base'
=== added file 'ivle/webapp/base/__init__.py'
=== added file 'ivle/webapp/base/ivle-headings.html'
--- ivle/webapp/base/ivle-headings.html	1970-01-01 00:00:00 +0000
+++ ivle/webapp/base/ivle-headings.html	2009-02-04 04:46:57 +0000
@@ -0,0 +1,78 @@
+<html xmlns="http://www.w3.org/1999/xhtml"       
+      xmlns:py="http://genshi.edgewall.org/"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      py:strip="">
+  <py:match path="head[@status!='done']" once="true">
+    <head status='done' py:attrs="select('@*')">
+      <title py:with="title = list(select('title/text()'))">
+        <py:if test="title">${title} - </py:if>IVLE
+      </title>
+
+      <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
+
+      <script py:if="write_javascript_settings" type="text/javascript">
+        root_dir = "${root_dir}";
+        public_host = "${public_host}";
+        username = "${login}";
+      </script>
+
+      <link py:if="favicon" rel="shortcut icon" href="${favicon}" />
+
+      <link rel="stylesheet" type="text/css" href="/media/common/ivle.css" />
+
+      <script type="text/javascript" src="/media/common/util.js"></script>
+      <script type="text/javascript" src="/media/common/json2.js"></script>
+      <script type="text/javascript" src="/media/common/md5.js"></script>
+      <script type="text/javascript" src="/media/common/tos.js"></script>
+
+      <!-- Plugin style sheets and JavaScript -->
+      <py:for each="style in app_styles">
+        <link rel="stylesheet" type="text/css" href="${style}" />
+      </py:for>
+
+      <py:for each="script in scripts">
+        <script type="text/javascript" src="${script}"></script>
+      </py:for>
+
+      <!-- Plugin JavaScript to be run on page load -->
+      <script type="text/javascript">
+        <py:for each="init_script in scripts_init">
+          window.addEventListener("load", ${init_script}, false);
+        </py:for>
+      </script>
+      ${select('*[local-name()!="title"]')}
+    </head>
+  </py:match>    
+  <py:match path="body[@status!='done']" once="true">
+    <body status='done'>
+      <div id="ivleheader"></div>
+      <div id="ivleheader_text">
+        <h1>IVLE</h1>
+        <h2>Informatics Virtual Learning Environment</h2>
+        <py:choose>
+          <p py:when="publicmode" class="userhello">Running in public mode.</p>
+          <p py:when="logged_in" class="userhello">
+            <span id="usernick">${nick}</span>
+            (<span class="username">${login}</span>) |
+            <a href="/settings">Settings</a> |
+            <a href="/help">Help</a> |
+            <a href="/logout">Sign out</a>
+          </p>
+          <p py:otherwise="">Not logged in.</p>
+        </py:choose>
+      </div>
+      <div id="ivleheader_tabs">
+        <ul id="apptabs">
+          <py:for each="app in apps_in_tabs">
+            <li py:if="app['this_app']" class="thisapp"><py:if test="app['has_icon']"><img src="${app['icon_url']}" alt="" /> </py:if><a href="${app['path']}" title="${app['desc']}">${app['name']}</a></li>
+            <li py:if="not app['this_app']"><py:if test="app['has_icon']"><img src="${app['icon_url']}" alt="" /> </py:if><a href="${app['path']}" title="${app['desc']}">${app['name']}</a></li>
+          </py:for>
+        </ul>
+      </div>
+      <div id="ivlebody">
+      ${select('*')}
+      </div>
+    </body>
+  </py:match>
+  ${app_template}
+</html>

=== added file 'ivle/webapp/base/plugins.py'
--- ivle/webapp/base/plugins.py	1970-01-01 00:00:00 +0000
+++ ivle/webapp/base/plugins.py	2009-02-03 13:02:36 +0000
@@ -0,0 +1,24 @@
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+
+# Author: Matt Giuca, Will Grant
+
+class BasePlugin(object):
+    """
+    Base class for all Plugin classes.
+    """
+    pass

=== added file 'ivle/webapp/base/test.py'
--- ivle/webapp/base/test.py	1970-01-01 00:00:00 +0000
+++ ivle/webapp/base/test.py	2009-02-04 05:06:26 +0000
@@ -0,0 +1,143 @@
+import urllib
+
+from ivle.webapp.base.views import RESTView, JSONRESTView, named_operation
+from ivle.webapp.errors import BadRequest, MethodNotAllowed
+from ivle.webapp.testing import FakeUser, FakeRequest
+
+class JSONRESTViewTest(JSONRESTView):
+    '''A small JSON REST view for testing purposes.'''
+    def GET(self, req):
+        return {'method': 'get'}
+
+    def PUT(self, req, data):
+        return {'method': 'put',
+                'result': data['result'], 'test': data['test']}
+
+    def PATCH(self, req, data):
+        return {'method': 'patch',
+                'result': data['result'], 'test': data['test']}
+
+    @named_operation
+    def do_stuff(self, what):
+        return {'result': 'Did %s!' % what}
+
+class TestJSONRESTView:
+    def testGET(self):
+        req = FakeRequest()
+        view = JSONRESTViewTest(req)
+        view.render(req)
+        assert req.content_type == 'application/json'
+        assert req.response_body == '{"method": "get"}\n'
+
+    def testPUT(self):
+        req = FakeRequest()
+        req.method = 'PUT'
+        req.request_body = '{"test": "FAI\\uA746ED", "result": 1}'
+        view = JSONRESTViewTest(req)
+        view.render(req)
+        assert req.content_type == 'application/json'
+        assert req.response_body == \
+                '{"test": "FAI\\ua746ED", "method": "put", "result": 1}\n'
+
+    def testPATCH(self):
+        req = FakeRequest()
+        req.method = 'PATCH'
+        req.request_body = '{"test": "FAI\\uA746ED", "result": 1}'
+        view = JSONRESTViewTest(req)
+        view.render(req)
+        assert req.content_type == 'application/json'
+        assert req.response_body == \
+                '{"test": "FAI\\ua746ED", "method": "patch", "result": 1}\n'
+
+    def testEmulatedPATCH(self):
+        req = FakeRequest()
+        req.method = 'PUT'
+        req.headers_in['X-IVLE-Patch-Semantics'] = 'yes'
+        req.request_body = '{"test": "FAI\\uA746ED", "result": 1}'
+        view = JSONRESTViewTest(req)
+        view.render(req)
+        assert req.content_type == 'application/json'
+        assert req.response_body == \
+                '{"test": "FAI\\ua746ED", "method": "patch", "result": 1}\n'
+
+    def testInvalidMethod(self):
+        req = FakeRequest()
+        req.method = 'FAKEANDBOGUS'
+        view = JSONRESTViewTest(req)
+        try:
+            view.render(req)
+        except MethodNotAllowed:
+            pass
+        else:
+            raise AssertionError("did not raise MethodNotAllowed")
+
+    def testInvalidMethodWithPATCHEmulation(self):
+        req = FakeRequest()
+        req.method = 'FAKEANDBOGUS'
+        req.headers_in['X-IVLE-Patch-Semantics'] = 'yes'
+        view = JSONRESTViewTest(req)
+        try:
+            view.render(req)
+        except MethodNotAllowed:
+            pass
+        else:
+            raise AssertionError("did not raise MethodNotAllowed")
+
+    def testNamedOperation(self):
+        req = FakeRequest()
+        req.method = 'POST'
+        req.request_body = urllib.urlencode({'ivle.op': 'do_stuff',
+                                             'what': 'blah'})
+        view = JSONRESTViewTest(req)
+        view.render(req)
+        assert req.content_type == 'application/json'
+        assert req.response_body == '{"result": "Did blah!"}\n'
+
+    def testPOSTWithoutName(self):
+        req = FakeRequest()
+        req.method = 'POST'
+        req.request_body = urllib.urlencode({'what': 'blah'})
+        view = JSONRESTViewTest(req)
+        try:
+            view.render(req)
+        except BadRequest, e:
+            assert e.message == 'No named operation specified.'
+        else:
+            raise AssertionError("did not raise BadRequest")
+
+    def testNonexistentNamedOperation(self):
+        req = FakeRequest()
+        req.method = 'POST'
+        req.request_body = urllib.urlencode({'ivle.op': 'enoent'})
+        view = JSONRESTViewTest(req)
+        try:
+            view.render(req)
+        except BadRequest, e:
+            assert e.message == 'Invalid named operation.'
+        else:
+            raise AssertionError("did not raise BadRequest")
+
+    def testDisallowedNamedOperation(self):
+        req = FakeRequest()
+        req.method = 'POST'
+        req.request_body = urllib.urlencode({'ivle.op': 'GET'})
+        view = JSONRESTViewTest(req)
+        try:
+            view.render(req)
+        except BadRequest, e:
+            assert e.message == 'Invalid named operation.'
+        else:
+            raise AssertionError("did not raise BadRequest")
+
+    def testNamedOperationWithMissingArgs(self):
+        req = FakeRequest()
+        req.method = 'POST'
+        req.request_body = urllib.urlencode({'ivle.op': 'do_stuff',
+                                             'nothing': 'wrong'})
+        view = JSONRESTViewTest(req)
+        try:
+            view.render(req)
+        except BadRequest, e:
+            assert e.message == 'Missing arguments: what'
+        else:
+            raise AssertionError("did not raise BadRequest")

=== added file 'ivle/webapp/base/views.py'
--- ivle/webapp/base/views.py	1970-01-01 00:00:00 +0000
+++ ivle/webapp/base/views.py	2009-02-04 06:37:54 +0000
@@ -0,0 +1,175 @@
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+
+# Author: Matt Giuca, Will Grant
+
+import inspect
+import cgi
+import os.path
+
+import cjson
+import genshi.template
+
+import ivle.conf
+import ivle.util
+from ivle.webapp.errors import BadRequest, MethodNotAllowed
+
+class BaseView(object):
+    """
+    Abstract base class for all view objects.
+    """
+    def __init__(self, req, **kwargs):
+        pass
+    def render(self, req):
+        pass
+
+class RESTView(BaseView):
+    """
+    A view which provides a RESTful interface. The content type is
+    unspecified (see JSONRESTView for a specific content type).
+    """
+    content_type = "application/octet-stream"
+
+    def __init__(self, req, *args, **kwargs):
+        pass
+
+    def render(self, req):
+        if req.method == 'GET':
+            outstr = self.GET(req)
+        # XXX PATCH hack
+        if req.method == 'PUT':
+            outstr = self.PATCH(req, req.read())
+        req.content_type = self.content_type
+        req.write(outstr)
+
+class JSONRESTView(RESTView):
+    """
+    A special case of RESTView which deals entirely in JSON.
+    """
+    content_type = "application/json"
+
+    def render(self, req):
+        if req.method == 'GET':
+            outjson = self.GET(req)
+        # Since PATCH isn't yet an official HTTP method, we allow users to
+        # turn a PUT into a PATCH by supplying a special header.
+        elif req.method == 'PATCH' or (req.method == 'PUT' and
+              'X-IVLE-Patch-Semantics' in req.headers_in and
+              req.headers_in['X-IVLE-Patch-Semantics'].lower() == 'yes'):
+            outjson = self.PATCH(req, cjson.decode(req.read()))
+        elif req.method == 'PUT':
+            outjson = self.PUT(req, cjson.decode(req.read()))
+        # POST implies named operation.
+        elif req.method == 'POST':
+            # TODO: Check Content-Type and implement multipart/form-data.
+            opargs = dict(cgi.parse_qsl(req.read()))
+            try:
+                opname = opargs['ivle.op']
+                del opargs['ivle.op']
+            except KeyError:
+                raise BadRequest('No named operation specified.')
+
+            try:
+                op = getattr(self, opname)
+            except AttributeError:
+                raise BadRequest('Invalid named operation.')
+
+            if not hasattr(op, '_rest_api_callable') or \
+               not op._rest_api_callable:
+                raise BadRequest('Invalid named operation.')
+
+            # Find any missing arguments, except for the first one (self).
+            argspec = inspect.getargspec(op)
+            missing = frozenset(argspec[0][1:]) - frozenset(opargs.keys())
+            if missing:
+                raise BadRequest('Missing arguments: ' + ', '.join(missing))
+
+            outjson = op(**opargs)
+        else:
+            raise MethodNotAllowed()
+
+        req.content_type = self.content_type
+        if outjson is not None:
+            req.write(cjson.encode(outjson))
+            req.write("\n")
+
+def named_operation(meth):
+    '''Declare a function to be accessible to HTTP users via the REST API.
+    '''
+    meth._rest_api_callable = True
+    return meth
+
+class XHTMLView(BaseView):
+    """
+    A view which provides a base class for views which need to return XHTML
+    It is expected that apps which use this view will be written using Genshi
+    templates.
+    """
+    def __init__(self, req, **kwargs):
+        for key in kwargs:
+          setattr(self, key, kwargs[key])
+        
+    def render(self, req):
+        req.content_type = 'text/html' # TODO: Detect application/xhtml+xml
+        ctx = genshi.template.Context()
+        self.populate(req, ctx)
+        self.populate_headings(req, ctx)
+        
+        ctx['app_styles'] = req.styles
+        ctx['scripts'] = req.scripts
+        ctx['scripts_init'] = req.scripts_init
+        app_template = os.path.join(os.path.dirname(
+                        inspect.getmodule(self).__file__), self.app_template) 
+        req.write_html_head_foot = False
+        loader = genshi.template.TemplateLoader(".", auto_reload=True)
+        tmpl = loader.load(app_template)
+        app = tmpl.generate(ctx)
+        ctx['app_template'] = app
+        tmpl = loader.load(os.path.join(os.path.dirname(__file__), 
+                                                        'ivle-headings.html'))
+        req.write(tmpl.generate(ctx).render('xhtml', doctype='xhtml'))
+
+    def populate_headings(self, req, ctx):
+        ctx['favicon'] = None
+        ctx['root_dir'] = ivle.conf.root_dir
+        ctx['public_host'] = ivle.conf.public_host
+        ctx['write_javascript_settings'] = req.write_javascript_settings
+        if req.user:
+            ctx['login'] = req.user.login
+            ctx['logged_in'] = True
+            ctx['nick'] = req.user.nick
+        else:
+            ctx['login'] = None
+        ctx['publicmode'] = req.publicmode
+        ctx['apps_in_tabs'] = []
+        for urlname in ivle.conf.apps.apps_in_tabs:
+            new_app = {}
+            app = ivle.conf.apps.app_url[urlname]
+            new_app['this_app'] = urlname == self.appname
+            if app.icon:
+                new_app['has_icon'] = True
+                icon_dir = ivle.conf.apps.app_icon_dir
+                icon_url = ivle.util.make_path(os.path.join(icon_dir, app.icon))
+                new_app['icon_url'] = icon_url
+                if new_app['this_app']:
+                    ctx['favicon'] = icon_url
+            else:
+                new_app['has_icon'] = False
+            new_app['path'] = ivle.util.make_path(urlname)
+            new_app['desc'] = app.desc
+            new_app['name'] = app.name
+            ctx['apps_in_tabs'].append(new_app)

=== renamed directory 'www/apps/browser' => 'ivle/webapp/browser'
=== modified file 'ivle/webapp/browser/__init__.py'
--- www/apps/browser/__init__.py	2009-01-19 22:56:46 +0000
+++ ivle/webapp/browser/__init__.py	2009-02-04 06:37:54 +0000
@@ -1,5 +1,5 @@
-# IVLE
-# Copyright (C) 2007-2008 The University of Melbourne
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -15,15 +15,19 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-# App: File Browser
-# Author: Matt Giuca
-# Date: 9/1/2008
-
-# The file browser application. Presents an Ajax-based interface to the
-# student's subversion workspace.
-# Note that there is virtually no server-side code for this application. It
-# simply presents static HTML and JavaScript, and all server-side activities
-# take place in the FileService app (for handling Ajax requests).
+# Author: Matt Giuca, Nick Chadwick
+
+"""
+The file browser application. Presents an Ajax-based interface to the
+student's subversion workspace.
+
+Note that there is virtually no server-side code for this application. It
+simply presents static HTML and JavaScript, and all server-side activities
+take place in the FileService app (for handling Ajax requests).
+"""
+
+from ivle.webapp.base.plugins import BasePlugin
+from ivle.webapp.base.views import XHTMLView
 
 import os.path
 import cgi
@@ -31,236 +35,143 @@
 from ivle import (util, studpath)
 import ivle.svn
 
-# url path for this app
-THIS_APP = "files"
-
-def handle(req):
-    """Handler for the File Browser application."""
-
-    # Work out where we are browsing
-    browsepath = req.path
-    if len(browsepath) == 0:
-        # If no path specified, default to the user's home directory
-        redirectPath = util.make_path(os.path.join(THIS_APP,req.user.login))
-        req.throw_redirect(util.make_path(redirectPath))
-
-    # Set request attributes
-    req.content_type = "text/html"
-    req.styles = [
-        "media/browser/browser.css",
-        "media/browser/listing.css",
-        "media/browser/editor.css",
-    ]
-    req.scripts = [
-        "media/common/json2.js",
-        "media/common/codepress/codepress.js",
-        "media/common/util.js",
-        "media/browser/browser.js",
-        "media/browser/listing.js",
-        "media/browser/editor.js",
-        "media/browser/specialhome.js",
-    ]
-    req.scripts_init = [
-        "browser_init",
-    ]
-
-    req.write_html_head_foot = True     # Have dispatch print head and foot
-    # The page title should contain the name of the file being browsed
-    req.title = browsepath.rsplit('/', 1)[-1]
-
-    _, localpath = studpath.url_to_local(browsepath)
-    if localpath is None:
-        req.throw_error(req.HTTP_NOT_FOUND,
-            "The path specified is invalid.")
-
-    # Start writing data
-    req.write("""
-<!-- Top bar section -->
-
-<div id="topbar">
-  <div id="path">
-    """)
-    # FIXME: This isn't completely reliable! We're not inside the jail, so we
-    # can't know the type for sure. This is now only used for adding a / to the
-    # end of displayed paths, so I'm leaving this although it will often break.
-    isdir = os.path.isdir(localpath)
-    presentpath(req, browsepath, isdir)
-    req.write("""
-  </div>
-  <div id="actions1">
-""")
-    present_actions1(req)
-    req.write("""  </div>
-  <div id="actions2">
-""")
-    present_actions2(req)
-    req.write("""  </div>
-</div>
-<!-- End topbar -->
-
-<!-- Body. The JavaScript places content here relevant to the path -->
-<div id="filesbody">
-</div>
-<!-- End body -->
-""")
-
-def presentpath(req, path, isdir):
-    """
-    Presents a path list (address bar inside the page) for clicking.
-    Writes to req, expecting to have just written the opening div containing
-    the listing.
-    This will also have a revision indicator on the end, if we are viewing a
-    revision.
-    """
-    href_path = util.make_path(THIS_APP)
-    nav_path = ""
-
-    revision = ivle.svn.revision_from_string(
-                     req.get_fieldstorage().getfirst('r'))
-
-    try: 
-        revno = revision.number
-    except:
-        revno = None
-       
-    # Create all of the paths
-    pathlist = path.split("/")
-    segs_left = len(pathlist)
-    for path_seg in pathlist:
-        if path_seg == "":
-            continue
-        # Write a slash at the end unless this is the last path seg AND
-        # it's not a directory.
-        segs_left -= 1
-        add_slash = segs_left != 0 or isdir
-        # Make an 'a' element
-        href_path = href_path + '/' + path_seg
-        nav_path = nav_path + path_seg
-        if add_slash:
-            nav_path = nav_path + '/'
-        link = '<a href="%s" title="Navigate to %s">%s</a>' % (
-            href_path if revno is None else href_path + '?r=%d' % revno,
-            nav_path, path_seg)
-        req.write(link)
-        if add_slash:
-            req.write('/')
-
-    if revno is not None:
-        req.write(' (revision %d)' % revno)
-
-def present_actions1(req):
-    """
-    Presents a set of links/buttons for the "actions1" row of the top bar.
-    This is always exactly the same - the JavaScript will customize it later.
-    """
-    # Set up our actions. The second field of each group is whether to disable
-    # the items by default.
-    moreactions = [
-      ('Publishing', True, [
-        ('publish', ['Publish',          'Make it so this directory can be seen by anyone on the web']),
-        ('share',   ['Share this file',  'Get a link to this published file, to give to friends']),
-        ('submit',  ['Submit', 'Submit the selected files for an assignment'])
-      ]),
-      ('File actions', True, [
-        ('rename',  ['Rename',           'Change the name of this file']),
-        ('delete',  ['Delete',           'Delete the selected files']),
-        ('copy',    ['Copy',             'Prepare to copy the selected files to another directory']),
-        ('cut',     ['Cut',              'Prepare to move the selected files to another directory'])
-      ]),
-      ('Directory actions', False, [
-        ('paste',   ['Paste',            'Paste the copied or cut files into the current directory']),
-        ('newfile', ['New File',         'Open a new file for editing in the current directory']),
-        ('mkdir',   ['New Directory',    'Make a new subdirectory in the current directory']),
-        ('upload',  ['Upload File',      'Upload a file to the current directory'])
-      ]),
-      ('Subversion', True, [
-        ('svncut',      ['Svn Cut',      'Prepare to move the selected files to another directory, maintaining history']),
-        ('svncopy',     ['Svn Copy',     'Prepare to copy the selected files to another directory, maintaining history']),
-        ('svnadd',      ['Add',            'Schedule the selected temporary files to be added permanently']),
-        ('svnremove',   ['Remove',         'Schedule the selected permanent files to be removed']),
-        ('svndiff',     ['Diff',           'View any changes to the selected file since its last committed state']),
-        ('svnrevert',   ['Revert',         'Restore the selected files back to their last committed state']),
-        ('svnupdate',   ['Update',         'Update your files with changes from the permanent repository']),
-        ('svncommit',   ['Commit',         'Commit any changes to the permanent repository']),
-        ('svnresolved', ['Mark Resolved',  'Mark a conflicted file as being resolved']),
-        ('svnlog',      ['View Log',       'View the log of commits of the selected file']),
-      ])
-    ]
-
-    req.write("""    <a id="act_open" class="disabled">Open</a> :
-    <a id="act_serve"
-        title="View this file on the web, running your code if this is a CGI file"
-        class="disabled" target="_blank">Serve</a> :
-    <a id="act_run" title="Run the selected Python file in the console"
-        class="disabled">Run</a> :
-    <a id="act_download" class="choice">Download</a> :
-    <a title="Refresh the current page" onclick="refresh()"
-        class="choice">Refresh</a><span id="moreactions_area"> :
-    <select id="moreactions" onchange="handle_moreactions()"
-        onblur="handle_moreactions()">
-      <option class="moreactions" value="top"
-        selected="selected">More actions...</option>
-""")
-
-    for (name, disablement, group) in moreactions:
-        if disablement:
-            disable = 'class="disabled" disabled="disabled"'
-        else:
-            disable = ''
-        req.write('<optgroup label="%s">' % name)
-        for (id, bits) in group:
-            req.write('<option id="act_%s" %s title="%s" value="%s">%s</option>'
-                      % (id, disable, bits[1], id, bits[0]))
-        req.write('</optgroup>')
-
-    req.write('</select></span>')
-
-def present_actions2(req):
-    """
-    Presents a set of links/buttons for the "actions2" row of the top bar.
-    This depends on whether it is a directory (listing) or normal file
-    (editor), but we'll let the JavaScript decide which.
-    """
-    req.write("""    <form id="actions2_directory"
-          target="upload_iframe"
-          action="%s"
-          enctype="multipart/form-data" method="post">
-      <p>
-      <span id="selectallpanel">
-      Select:
-      <a onclick="action_selectall(true)"
-          title="Select all files in this directory">All</a> :
-      <a onclick="action_selectall(false)"
-          title="Deselect all files in this directory">None</a>
-      </span>
-
-      <span style="display:none" id="uploadpanel">|
-        <label for="upload_file">Upload file:</label>
-        <input type="hidden" value="putfiles" name="action" />
-        <input type="hidden" value="" name="path" />
-        <input type="file" name="data" id="upload_file" />
-        <input type="checkbox" checked="checked" value="true" name="unpack" id="unpack" />
-        <label title="Extract files and directories from a ZIP file if one is uploaded" for="unpack">Unpack ZIP</label>
-        <input type="button" onclick="show_uploadpanel(false)" value="Cancel" />
-        <input type="submit" value="Upload" />
-      </span>
-      <!-- This iframe is for making a call to upload the file without
-           refreshing the page. (It will refresh the listing). -->
-      <iframe onload="upload_callback()" style="display: none;"
-          name="upload_iframe" id="upload_iframe"></iframe>
-      </p>
-    </form>
-""" % cgi.escape(util.make_path(os.path.join("fileservice", req.path))))
-
-    req.write("""    <p id="actions2_file">
-      <input type="button" id="save_button" value="Save" onclick="save_file('%s')" />
-      <input type="button" id="saveas_button" value="Save As..." onclick="save_file_as('%s')" />
-      <select id="highlighting_select" onchange="highlighting_changed(this)">
-          <option value="text">Text</option>
-          <option value="python">Python</option>
-          <option value="html">HTML</option>
-          <option value="javascript">JavaScript</option>
-          <option value="css">CSS</option>
-      </select>
-    </p>
-""" % ((cgi.escape(req.path),) * 2))
+class BrowserView(XHTMLView):
+    """
+    The view for the browser
+    """
+    def __init__(self, req, **kwargs):
+        super(BrowserView, self).__init__(req, **kwargs)
+        self.app_template = 'template.html'
+        #XXX: Will hate me for this.
+        self.appname = "files"
+
+    def populate(self, req, ctx):
+        if not hasattr(self, 'path'):
+            # If no path specified, default to the user's home directory
+            redirectPath = util.make_path(os.path.join('files', req.user.login))
+            req.throw_redirect(util.make_path(redirectPath))
+
+        # Set request attributes
+        req.styles = [
+            "/media/browser/browser.css",
+            "/media/browser/listing.css",
+            "/media/browser/editor.css",
+        ]
+        req.scripts = [
+            "/media/common/json2.js",
+            "/media/common/codepress/codepress.js",
+            "/media/common/util.js",
+            "/media/browser/browser.js",
+            "/media/browser/listing.js",
+            "/media/browser/editor.js",
+            "/media/browser/specialhome.js",
+        ]
+        req.scripts_init = [
+            "browser_init",
+        ]
+
+        # The page title should contain the name of the file being browsed
+        req.title = self.path.rsplit('/', 1)[-1]
+
+        _, localpath = studpath.url_to_local(self.path)
+        if localpath is None:
+            req.throw_error(req.HTTP_NOT_FOUND,
+                "The path specified is invalid.")
+
+        # Start writing data
+
+        # FIXME: This isn't completely reliable! We're not inside the jail, so we
+        # can't know the type for sure. This is now only used for adding a / to the
+        # end of displayed paths, so I'm leaving this although it will often break.
+        isdir = os.path.isdir(localpath)
+        ctx['isdir'] = isdir
+        self.gen_path(req, ctx)
+        self.gen_actions(req, ctx)
+
+        ctx['fileservice_action'] = util.make_path(os.path.join("fileservice", req.path))
+        ctx['filename'] = cgi.escape(req.path)
+
+    #TODO: Move all this logic into the template
+    def gen_path(self, req, ctx):
+
+        href_path = util.make_path('files')
+        nav_path = ""
+        revision = ivle.svn.revision_from_string(
+                         req.get_fieldstorage().getfirst('r'))
+        try:
+            revno = revision.number
+        except:
+            revno = None
+
+        ctx['revno'] = revno
+
+        # Create all of the paths
+        pathlist = self.path.split("/")
+        ctx['paths'] = []
+        for path_seg in pathlist:
+            if path_seg == "":
+                continue
+            new_seg = {}
+
+            nav_path = nav_path + path_seg
+            href_path = href_path + '/' + path_seg
+
+            new_seg['path'] = path_seg
+            new_seg['nav_path'] = nav_path
+            new_seg['href_path'] = href_path
+            if revno is not None:
+                new_seg['href_path'] += '?r=%d' % revno
+
+            ctx['paths'].append(new_seg)
+
+    def gen_actions(self, req, ctx):
+        """
+        Presents a set of links/buttons for the "actions1" row of the top bar.
+        This is always exactly the same - the JavaScript will customize it later.
+        """
+        # Set up our actions. The second field of each group is whether to disable
+        # the items by default.
+        ctx['moreactions'] = [
+          ('Publishing', True, [
+            ('publish', ['Publish',          'Make it so this directory can be seen by anyone on the web']),
+            ('share',   ['Share this file',  'Get a link to this published file, to give to friends']),
+            ('submit',  ['Submit', 'Submit the selected files for an assignment'])
+          ]),
+          ('File actions', True, [
+            ('rename',  ['Rename',           'Change the name of this file']),
+            ('delete',  ['Delete',           'Delete the selected files']),
+            ('copy',    ['Copy',             'Prepare to copy the selected files to another directory']),
+            ('cut',     ['Cut',              'Prepare to move the selected files to another directory'])
+          ]),
+          ('Directory actions', False, [
+            ('paste',   ['Paste',            'Paste the copied or cut files into the current directory']),
+            ('newfile', ['New File',         'Open a new file for editing in the current directory']),
+            ('mkdir',   ['New Directory',    'Make a new subdirectory in the current directory']),
+            ('upload',  ['Upload File',      'Upload a file to the current directory'])
+          ]),
+          ('Subversion', True, [
+            ('svncut',      ['Svn Cut',      'Prepare to move the selected files to another directory, maintaining history']),
+            ('svncopy',     ['Svn Copy',     'Prepare to copy the selected files to another directory, maintaining history']),
+            ('svnadd',      ['Add',            'Schedule the selected temporary files to be added permanently']),
+            ('svnremove',   ['Remove',         'Schedule the selected permanent files to be removed']),
+            ('svndiff',     ['Diff',           'View any changes to the selected file since its last committed state']),
+            ('svnrevert',   ['Revert',         'Restore the selected files back to their last committed state']),
+            ('svnupdate',   ['Update',         'Update your files with changes from the permanent repository']),
+            ('svncommit',   ['Commit',         'Commit any changes to the permanent repository']),
+            ('svnresolved', ['Mark Resolved',  'Mark a conflicted file as being resolved']),
+            ('svnlog',      ['View Log',       'View the log of commits of the selected file']),
+          ])
+        ]
+
+class Plugin(BasePlugin):
+    """
+    The Plugin class for the user plugin.
+    """
+    # Magic attribute: urls
+    # Sequence of pairs/triples of
+    # (regex str, handler class, kwargs dict)
+    # The kwargs dict is passed to the __init__ of the view object
+    urls = [
+        ('files/*(path)', BrowserView),
+        ('files/', BrowserView),
+    ]

=== added file 'ivle/webapp/browser/template.html'
--- ivle/webapp/browser/template.html	1970-01-01 00:00:00 +0000
+++ ivle/webapp/browser/template.html	2009-02-04 03:14:31 +0000
@@ -0,0 +1,104 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:py="http://genshi.edgewall.org/">
+<!-- Top bar section -->
+<head>
+  <title>File Browser</title>
+</head>
+<body>
+  <div id="topbar">
+    <div id="path">
+      <span>
+        <py:for each="segment in paths">/<a href="${segment['href_path']}" title="Navigate to ${segment['nav_path']}">${segment['path']}</a></py:for><py:if test="isdir">/</py:if><py:if test="revno is not None"> (revision ${revno})</py:if>
+      </span>
+    </div>
+    <div id="actions1">
+      <a id="act_open" class="disabled">Open</a> :
+      <a id="act_serve"
+          title="View this file on the web, running your code if this is a CGI file"
+          class="disabled" target="_blank">Serve</a> :
+      <a id="act_run" title="Run the selected Python file in the console"
+          class="disabled">Run</a> :
+      <a id="act_download" class="choice">Download</a> :
+      <a title="Refresh the current page" onclick="refresh()"
+          class="choice">Refresh</a>
+      <span id="moreactions_area"> :
+        <select id="moreactions" onchange="handle_moreactions()"
+            onblur="handle_moreactions()">
+          <option class="moreactions" value="top"
+              selected="selected">More actions...</option>
+          <py:for each="(name, disablement, group) in moreactions">
+
+            <optgroup label="${name}">
+
+              <py:choose test="disablement">
+
+                <py:when test="True">
+                  <py:for each="(id, bits) in group">
+                    <option id="act_${id}" class="disabled" disabled="disabled"
+                        title="${bits[1]}" value="${id}">${bits[0]}</option>
+                  </py:for>
+                </py:when>
+
+                <py:otherwise>
+                  <py:for each="(id, bits) in group">
+                    <option id="act_${id}" title="${bits[1]}" 
+                        value="${id}">${bits[0]}</option>
+                  </py:for>                
+                </py:otherwise>
+
+              </py:choose>
+
+            </optgroup>
+          </py:for>
+        </select>
+      </span>
+    </div>
+    <div id="actions2">
+      <form id="actions2_directory"
+          target="upload_iframe"
+          action="${fileservice_action}"
+          enctype="multipart/form-data" method="post">
+        <p>
+          <span id="selectallpanel">
+          Select:
+            <a onclick="action_selectall(true)"
+                title="Select all files in this directory">All</a> :
+            <a onclick="action_selectall(false)"
+                title="Deselect all files in this directory">None</a>
+          </span>
+
+          <span style="display:none" id="uploadpanel">|
+            <label for="upload_file">Upload file:</label>
+            <input type="hidden" value="putfiles" name="action" />
+            <input type="hidden" value="" name="path" />
+            <input type="file" name="data" id="upload_file" />
+            <input type="checkbox" checked="checked" value="true" name="unpack" id="unpack" />
+            <label title="Extract files and directories from a ZIP file if one is uploaded" for="unpack">Unpack ZIP</label>
+            <input type="button" onclick="show_uploadpanel(false)" value="Cancel" />
+            <input type="submit" value="Upload" />
+          </span>
+          <!-- This iframe is for making a call to upload the file without
+               refreshing the page. (It will refresh the listing). -->
+          <iframe onload="upload_callback()" style="display: none;"
+              name="upload_iframe" id="upload_iframe"></iframe>
+        </p>
+      </form>
+      <p id="actions2_file">
+        <input type="button" id="save_button" value="Save" onclick="save_file('${filename}')" />
+        <input type="button" id="saveas_button" value="Save As..." onclick="save_file_as('${filename}')" />
+        <select id="highlighting_select" onchange="highlighting_changed(this)">
+          <option value="text">Text</option>
+          <option value="python">Python</option>
+          <option value="html">HTML</option>
+          <option value="javascript">JavaScript</option>
+          <option value="css">CSS</option>
+        </select>
+      </p>
+    </div>
+  </div>
+  <!-- Body. The JavaScript places content here relevant to the path -->
+  <div id="filesbody">
+  </div>
+  <!-- End body -->
+</body>
+</html>

=== added file 'ivle/webapp/errors.py'
--- ivle/webapp/errors.py	1970-01-01 00:00:00 +0000
+++ ivle/webapp/errors.py	2009-02-04 05:06:26 +0000
@@ -0,0 +1,33 @@
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+
+# Author: Will Grant
+
+class HTTPError(Exception):
+    '''A base class for all HTTP errors.'''
+
+class BadRequest(HTTPError):
+    code = 400
+
+class Forbidden(HTTPError):
+    code = 403
+
+class NotFound(HTTPError):
+    code = 404
+
+class MethodNotAllowed(HTTPError):
+    code = 405

=== renamed directory 'www/apps/groups' => 'ivle/webapp/groups'
=== modified file 'ivle/webapp/groups/__init__.py'
--- www/apps/groups/__init__.py	2009-01-19 04:03:07 +0000
+++ ivle/webapp/groups/__init__.py	2009-02-04 06:37:54 +0000
@@ -27,84 +27,100 @@
 
 from ivle import caps
 from ivle.database import Subject
-
-def handle(req):
-    # Set request attributes
-    req.content_type = "text/html"
-    req.styles = ["media/groups/groups.css"]
-    req.scripts = [
-        "media/groups/groups.js",
-        "media/common/util.js",
-        "media/common/json2.js",
+from ivle import util
+
+from ivle.webapp.base.plugins import BasePlugin
+from ivle.webapp.base.views import XHTMLView
+
+class GroupsView(XHTMLView):
+    """
+    The groups view
+    """
+    def __init__(self, req, **kwargs):
+        super(GroupsView, self).__init__(req, **kwargs)
+        self.app_template = 'template.html'
+        #XXX: Will hates me for this.
+        self.appname = "groups"
+
+    def populate(self, req, ctx):
+        # Set request attributes
+        req.styles = ["media/groups/groups.css"]
+        req.scripts = [
+            "media/groups/groups.js",
+            "media/common/util.js",
+            "media/common/json2.js",
+        ]
+
+        ctx['enrolments'] = []
+        # Show a group panel per enrolment
+        enrolments = req.user.active_enrolments
+        if enrolments.count() == 0:
+            ctx['no_enrolments'] = True
+        else:
+            ctx['no_enrolments'] = False
+
+        for enrolment in enrolments:
+            self.add_subject_panel(req, enrolment.offering, ctx)
+
+        if req.user.hasCap(caps.CAP_MANAGEGROUPS):
+            ctx['manage_groups'] = True
+            ctx['manage_subjects'] = []
+            subjects = req.store.find(Subject)
+            for s in subjects:
+                new_s = {}
+                new_s['id'] = s.id
+                new_s['name'] = s.name
+                new_s['code'] = s.code
+                ctx['manage_subjects'].append(new_s)
+        else:
+            ctx['manage_groups'] = False
+
+
+    def add_subject_panel(self, req, offering, ctx):
+        """
+        Show the group management panel for a particular subject.
+        Prints to req.
+        """
+        # Get the groups this user is in, for this offering
+        groups = req.user.get_groups(offering)
+        if groups.count() == 0:
+            return
+
+        offering_groups = {}
+
+        offering_groups['offering_id'] = offering.id
+        offering_groups['offering_name'] = offering.subject.name
+        offering_groups['groups'] = []
+
+        #TODO: Use a better way to manage group membership and invitations
+        for group in groups:
+            new_group = {}
+            new_group['nick'] = cgi.escape(group.nick if group.nick else '')
+            new_group['name'] = cgi.escape(group.name)
+
+            # XXX - This should be set to reflect whether or not a user is invited
+            #     - or if they have accepted the offer
+            new_group['is_member'] = True
+            new_group['members'] = []
+
+            for user in group.members:
+                member = {}
+                member['fullname'] = cgi.escape(user.fullname)
+                member['login'] = cgi.escape(user.login)
+                new_group['members'].append(member)
+            offering_groups['groups'].append(new_group)
+
+        ctx['enrolments'].append(offering_groups)
+
+
+class Plugin(BasePlugin):
+    """
+    The Plugin class for the user plugin.
+    """
+    # Magic attribute: urls
+    # Sequence of pairs/triples of
+    # (regex str, handler class, kwargs dict)
+    # The kwargs dict is passed to the __init__ of the view object
+    urls = [
+        ('groups/', GroupsView),
     ]
-    req.write_html_head_foot = True     # Have dispatch print head and foot
-
-    req.write('<div id="ivle_padding">\n')
-
-    # Show a group panel per enrolment
-    enrolments = req.user.active_enrolments
-    if enrolments.count() == 0:
-        req.write("<p>Error: You are not currently enrolled in any subjects."
-                  "</p>\n")
-    for enrolment in enrolments:
-        show_subject_panel(req, enrolment.offering)
-    if req.user.hasCap(caps.CAP_MANAGEGROUPS):
-        show_groupadmin_panel(req)
-
-    req.write("</div>\n")
-
-def show_groupadmin_panel(req):
-    """
-    Shows the group admin panel
-    """
-    req.write("<hr/>\n")
-    req.write("<h1>Group Administration</h1>")
-    # Choose subject
-    subjects = req.store.find(Subject)
-    req.write("<label for=\"subject_select\">Subject:</label>\n")
-    req.write("<select id=\"subject_select\">\n")
-    for s in subjects:
-        req.write("    <option value=\"%d\">%s (%s)</option>\n"%
-            (s.id, s.name, s.code))
-    req.write("</select>\n")
-    req.write("<input type=\"button\" value=\"Manage\" \
-        onclick=\"manage_subject()\" />\n")
-    req.write("<div id=\"subject_div\"></div>")
-
-def show_subject_panel(req, offering):
-    """
-    Show the group management panel for a particular subject.
-    Prints to req.
-    """
-    # Get the groups this user is in, for this offering
-    groups = req.user.get_groups(offering)
-    if groups.count() == 0:
-        return
-
-    req.write("<div id=\"subject%d\"class=\"subject\">"%offering.id)
-    req.write("<h1>%s</h1>\n" % cgi.escape(offering.subject.name))
-    for group in groups:
-        req.write("<h2>%s (%s)</h2>\n" %
-            (cgi.escape(group.nick if group.nick else ''),
-             cgi.escape(group.name)))
-        if True: # XXX - was is_member (whether real member or just invited)
-            req.write('<p>You are a member of this group.</p>\n')
-        else:
-            req.write('<p>You have been invited to this group.</p>\n')
-            req.write('<p>'
-                '<input type="button" '
-                'onclick="accept(&quot;%(groupnm)s&quot;)" '
-                'value="Accept" />\n'
-                '<input type="button" '
-                'onclick="decline(&quot;%(groupnm)s&quot;)" '
-                'value="Decline" />\n'
-                '</p>\n' % {"groupnm": cgi.escape(group.name)})
-        req.write("<h3>Members</h3>\n")
-        req.write("<ul>\n")
-        for user in group.members:
-            req.write("<li>%s (%s)</li>" %
-                      (cgi.escape(user.fullname),
-                       cgi.escape(user.login)))
-        req.write("</ul>\n")
-
-    req.write("</div>")

=== added file 'ivle/webapp/groups/template.html'
--- ivle/webapp/groups/template.html	1970-01-01 00:00:00 +0000
+++ ivle/webapp/groups/template.html	2009-02-04 05:10:08 +0000
@@ -0,0 +1,54 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <head>
+    <title>Groupz</title>
+  </head>
+  <body>
+    <div id="ivle_padding">
+      <p py:if="no_enrolments">Error: You are not currently enrolled in any subjects.</p>
+      <py:for each="enrolment in enrolments">
+        <div id="subject${enrolment.offering_id}" class="subject">
+          <h1>${enrolment.offering_name}</h1>
+          <py:for each="group in enrolment['groups']">
+            <h2>${group['nick']} (${group['name']})</h2>
+            <py:choose test="group['is_member']">
+              <py:when test="True">
+                <p>You are a member of this group.</p>
+              </py:when>
+              <py:otherwise>
+                <p>You have been invited to this group.</p>
+                <p>
+                  <input type="button" 
+                      onclick="accept(&quot;${group['name']}&quot;)" 
+                      value="Accept" />
+                  <input type="button"
+                      onclick="decline(&quot;${group['name']}&quot;)"
+                      value="Decline" />
+                </p>
+              </py:otherwise>
+            </py:choose>
+            <h3>Members</h3>
+            <ul>
+              <py:for each="member in group['members']">
+                <li>${member['fullname']} (${member['login']})</li>
+              </py:for>
+            </ul>
+          </py:for>
+        </div>
+      </py:for>
+      <py:if test="manage_groups">
+        <hr />
+        <h1>Group Administration</h1>
+        <label for="subject_select">Subject:</label>
+        <select id="subject_select">
+          <py:for each="subject in manage_subjects">
+            <option value="${subject['id']}">${subject['name']} (${subject['code']})</option>
+          </py:for>
+        </select>
+        <input type="button" value="Manage" onclick="manage_subject()" />
+        <div id="subject_div"></div>
+      </py:if>
+    </div>
+  </body>
+</html>

=== added directory 'ivle/webapp/testing'
=== added file 'ivle/webapp/testing/__init__.py'
--- ivle/webapp/testing/__init__.py	1970-01-01 00:00:00 +0000
+++ ivle/webapp/testing/__init__.py	2009-02-04 02:00:21 +0000
@@ -0,0 +1,88 @@
+import datetime
+
+from ivle.database import User
+from ivle.dispatch.request import Request
+
+class FakeUser(User):
+    login = u'fakeuser'
+    state = u'enabled'
+    rolenm = u'student'
+    unixid = 5000
+    nick = u'Fake User'
+    pass_exp = None
+    acct_exp = None
+    last_login = datetime.datetime.now()
+    svn_pass = u'somepass'
+    email = u'fakeuser@example.com'
+    fullname = u'Fake Fixture User'
+    studentid = u'1234'
+
+class FakeRequest(Request):
+    '''A fake request object, for use as a fixture in tests.
+
+    This tries to behave fairly closely to an ivle.dispatch.request.Request,
+    but without needing a web server.
+    '''
+    def __init__(self):
+        '''Set sane defaults.'''
+
+        # Some fields are omitted because they make no sense in the new model.
+        self.headers_written = False
+        self.publicmode = False
+        self.method = 'GET'
+        self.uri = '/'
+        self.user = FakeUser()
+        self.hostname = 'fakehost'
+        self.headers_in = {}
+        self.headers_out = {}
+
+        # We don't want DB access in tests (by default)
+        self.store = None
+
+        # Default values for the output members
+        self.status = Request.HTTP_OK
+        self.content_type = None        # Use Apache's default
+        self.location = None
+        self.title = None     # Will be set by dispatch before passing to app
+        self.styles = []
+        self.scripts = []
+        self.scripts_init = []
+        self.write_html_head_foot = False
+        self.request_body = ''
+        self.response_body = ''
+        # In some cases we don't want the template JS (such as the username
+        # and public FQDN) in the output HTML. In that case, set this to 0.
+        self.write_javascript_settings = True
+        self.got_common_vars = False
+
+    def __del__(self):
+        '''Cleanup, but don't close the nonexistent store.'''
+        if self.store is not None:
+            self.store.close()
+
+    def ensure_headers_written(self):
+        '''Fake a write of the HTTP and HTML headers if they haven't already
+           been written.'''
+        pass
+
+    def read(self, len=None):
+        if len is None:
+            data = self.request_body
+            self.request_body = ''
+        else:
+            data = self.request_body[:len]
+            self.request_body = self.request_body[len:]
+        return data
+
+    def write(self, string, flush=1):
+        '''Write a string to the internal output storage.'''
+        self.ensure_headers_written()
+        if isinstance(string, unicode):
+            self.response_body += string.encode('utf8')
+        else:
+            self.response_body += string
+
+    def flush(self):
+        '''Fake a flush.'''
+        pass
+

=== renamed directory 'www/apps/tutorial' => 'ivle/webapp/tutorial'
=== modified file 'ivle/webapp/tutorial/__init__.py'
--- www/apps/tutorial/__init__.py	2009-01-27 05:13:56 +0000
+++ ivle/webapp/tutorial/__init__.py	2009-02-04 12:58:22 +0000
@@ -15,18 +15,13 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-# App: tutorial
-# Author: Matt Giuca
-# Date: 25/1/2008
-
-# Tutorial application.
-# Displays tutorial content with editable exercises, allowing students to test
-# and submit their solutions to exercises and have them auto-tested.
-
-# URL syntax
-# All path segments are optional (omitted path segments will show menus).
-# The first path segment is the subject code.
-# The second path segment is the worksheet name.
+# Author: Matt Giuca, Will Grant
+
+'''Tutorial/worksheet/exercise application.
+
+Displays tutorial content with editable exercises, allowing students to test
+and submit their solutions to exercises and have them auto-tested.
+'''
 
 import os
 import os.path
@@ -38,11 +33,16 @@
 import mimetypes
 
 import cjson
+import genshi
 
 from ivle import util
 import ivle.conf
 import ivle.database
+from ivle.database import Subject
 import ivle.worksheet
+from ivle.webapp.base.views import BaseView, XHTMLView
+from ivle.webapp.base.plugins import BasePlugin
+from ivle.webapp.errors import NotFound, Forbidden
 
 from rst import rst
 
@@ -56,6 +56,11 @@
         self.id = id
         self.name = name
         self.assessable = assessable
+        self.loc = urllib.quote(id)
+        self.complete_class = ''
+        self.optional_message = ''
+        self.total = 0
+        self.mand_done = 0
     def __repr__(self):
         return ("Worksheet(id=%s, name=%s, assessable=%s)"
                 % (repr(self.id), repr(self.name), repr(self.assessable)))
@@ -74,371 +79,247 @@
         else:
             return util.make_path(os.path.join(THIS_APP, subject, worksheet))
 
-def handle(req):
-    """Handler for the Tutorial application."""
-
-    # Set request attributes
-    req.content_type = "text/html"
-    req.scripts = [
-        "media/common/util.js",
-        "media/common/json2.js",
-        "media/tutorial/tutorial.js",
-    ]
-    req.styles = [
-        "media/tutorial/tutorial.css",
-    ]
-    # Note: Don't print write_html_head_foot just yet
-    # If we encounter errors later we do not want this
-
-    path_segs = req.path.split('/')
-    subject = None
-    worksheet = None
-    if len(req.path) > 0:
-        subject = path_segs[0]
-        if subject == "media":
-            # Special case: "tutorial/media" will plainly serve any path
-            # relative to "subjects/media".
-            handle_media_path(req)
-            return
-        if len(path_segs) > 2:
-            req.throw_error(req.HTTP_NOT_FOUND,
-                "Invalid tutorial path.")
-        if len(path_segs) == 2:
-            worksheet = path_segs[1]
-
-    if subject == None:
-        handle_toplevel_menu(req)
-    elif worksheet == None:
-        handle_subject_menu(req, subject)
-    else:
-        handle_worksheet(req, subject, worksheet)
-
-def handle_media_path(req):
-    """
-    Urls in "tutorial/media" will just be served directly, relative to
-    subjects. So if we came here, we just want to serve a file relative to the
-    subjects directory on the local file system.
-    """
-    # First normalise the path
-    urlpath = os.path.normpath(req.path)
-    # Now if it begins with ".." or separator, then it's illegal
-    if urlpath.startswith("..") or urlpath.startswith('/'):
-        req.throw_error(req.HTTP_FORBIDDEN,
-            "Invalid path.")
-    filename = os.path.join(ivle.conf.subjects_base, urlpath)
-    (type, _) = mimetypes.guess_type(filename)
-    if type is None:
-        type = ivle.conf.mimetypes.default_mimetype
-    ## THIS CODE taken from apps/server/__init__.py
-    if not os.access(filename, os.R_OK):
-        req.throw_error(req.HTTP_NOT_FOUND,
-            "The requested file does not exist.")
-    if os.path.isdir(filename):
-        req.throw_error(req.HTTP_FORBIDDEN,
-            "The requested file is a directory.")
-    req.content_type = type
-    req.sendfile(filename)
-
-def handle_toplevel_menu(req):
-    # This is represented as a directory. Redirect and add a slash if it is
-    # missing.
-    if req.uri[-1] != '/':
-        req.throw_redirect(make_tutorial_path())
-    req.write_html_head_foot = True
-    req.write('<div id="ivle_padding">\n')
-    req.write("<h1>IVLE Tutorials</h1>\n")
-    req.write("""<p>Welcome to the IVLE tutorial system.
-  Please select a subject from the list below to select a worksheet
-  for that subject.</p>\n""")
-
-    enrolled_subjects = req.user.subjects
-    unenrolled_subjects = [subject for subject in
-                           req.store.find(ivle.database.Subject)
-                           if subject not in enrolled_subjects]
-
-    def print_subject(subject):
-        req.write('  <li><a href="%s">%s</a></li>\n'
-            % (urllib.quote(subject.code) + '/',
-               cgi.escape(subject.name)))
-
-    req.write("<h2>Subjects</h2>\n<ul>\n")
-    for subject in enrolled_subjects:
-        print_subject(subject)
-    req.write("</ul>\n")
-    if len(unenrolled_subjects) > 0:
-        req.write("<h3>Other Subjects</h3>\n")
-        req.write("<p>You are not currently enrolled in these subjects.\n"
-                  "   Your marks will not be counted.</p>\n")
-        req.write("<ul>\n")
-        for subject in unenrolled_subjects:
-            print_subject(subject)
-        req.write("</ul>\n")
-    req.write("</div>\n")   # tutorialbody
+class SubjectView(XHTMLView):
+    '''The view of the index of worksheets for a subject.'''
+    app_template = 'subjectmenu.html'
+    appname = 'tutorial' # XXX
+
+    def __init__(self, req, subject):
+        self.subject = req.store.find(Subject, code=subject).one()
+
+    def populate(self, req, ctx):
+        if not self.subject:
+            raise NotFound()
+
+        # Subject names must be valid identifiers
+        if not is_valid_subjname(self.subject.code):
+            raise NotFound()
+
+        # Parse the subject description file
+        # The subject directory must have a file "subject.xml" in it,
+        # or it does not exist (404 error).
+        ctx['subject'] = self.subject.code
+        try:
+            subjectfile = open(os.path.join(ivle.conf.subjects_base,
+                                    self.subject.code, "subject.xml")).read()
+        except:
+            raise NotFound()
+
+        subjectfile = genshi.Stream(list(genshi.XML(subjectfile)))
+
+        ctx['worksheets'] = get_worksheets(subjectfile)
+
+        # As we go, calculate the total score for this subject
+        # (Assessable worksheets only, mandatory problems only)
+        problems_done = 0
+        problems_total = 0
+        for worksheet in ctx['worksheets']:
+            stored_worksheet = ivle.database.Worksheet.get_by_name(req.store,
+                self.subject.code, worksheet.id)
+            # If worksheet is not in database yet, we'll simply not display
+            # data about it yet (it should be added as soon as anyone visits
+            # the worksheet itself).
+            if stored_worksheet is not None:
+                # If the assessable status of this worksheet has changed,
+                # update the DB
+                # (Note: This fails the try block if the worksheet is not yet
+                # in the DB, which is fine. The author should visit the
+                # worksheet page to get it into the DB).
+                if worksheet.assessable != stored_worksheet.assessable:
+                    # XXX If statement to avoid unnecessary database writes.
+                    # Is this necessary, or will Storm check for us?
+                    stored_worksheet.assessable = worksheet.assessable
+                if worksheet.assessable:
+                    # Calculate the user's score for this worksheet
+                    mand_done, mand_total, opt_done, opt_total = (
+                        ivle.worksheet.calculate_score(req.store, req.user,
+                            stored_worksheet))
+                    if opt_total > 0:
+                        optional_message = " (excluding optional exercises)"
+                    else:
+                        optional_message = ""
+                    if mand_done >= mand_total:
+                        worksheet.complete_class = "complete"
+                    elif mand_done > 0:
+                        worksheet.complete_class = "semicomplete"
+                    else:
+                        worksheet.complete_class = "incomplete"
+                    problems_done += mand_done
+                    problems_total += mand_total
+                    worksheet.mand_done = mand_done
+                    worksheet.total = mand_total
+                    worksheet.optional_message = optional_message
+
+
+        ctx['problems_total'] = problems_total
+        ctx['problems_done'] = problems_done
+        if problems_total > 0:
+            if problems_done >= problems_total:
+                ctx['complete_class'] = "complete"
+            elif problems_done > 0:
+                ctx['complete_class'] = "semicomplete"
+            else:
+                ctx['complete_class'] = "incomplete"
+            ctx['problems_pct'] = (100 * problems_done) / problems_total
+            # TODO: Put this somewhere else! What is this on about? Why 16?
+            # XXX Marks calculation (should be abstracted out of here!)
+            # percent / 16, rounded down, with a maximum mark of 5
+            ctx['max_mark'] = 5
+            ctx['mark'] = min(ctx['problems_pct'] / 16, ctx['max_mark'])
+
+class WorksheetView(XHTMLView):
+    '''The view of a worksheet with exercises.'''
+    app_template = 'worksheet.html'
+    appname = 'tutorial' # XXX
+
+    def __init__(self, req, subject, worksheet):
+        self.subject = req.store.find(Subject, code=subject).one()
+        self.worksheetname = worksheet
+
+    def populate(self, req, ctx):
+        req.scripts = [
+            "/media/common/util.js",
+            "/media/common/json2.js",
+            "/media/tutorial/tutorial.js",
+        ]
+        req.styles = [
+            "/media/tutorial/tutorial.css",
+        ]
+
+        if not self.subject:
+            raise NotFound()
+
+        # Subject and worksheet names must be valid identifiers
+        if not is_valid_subjname(self.subject.code) or \
+           not is_valid_subjname(self.worksheetname):
+            raise NotFound()
+
+        # Read in worksheet data
+        worksheetfilename = os.path.join(ivle.conf.subjects_base,
+                               self.subject.code, self.worksheetname + ".xml")
+        try:
+            worksheetfile = open(worksheetfilename)
+            worksheetmtime = os.path.getmtime(worksheetfilename)
+        except:
+            raise NotFound()
+
+        worksheetmtime = datetime.fromtimestamp(worksheetmtime)
+        worksheetfile = worksheetfile.read()
+
+        ctx['subject'] = self.subject.code
+        ctx['worksheetstream'] = genshi.Stream(list(genshi.XML(worksheetfile)))
+
+        #TODO: Replace this with a nice way, possibly a match template
+        generate_worksheet_data(ctx, req)
+
+        update_db_worksheet(req.store, self.subject.code, self.worksheetname,
+            worksheetmtime, ctx['exerciselist'])
+
+        ctx['worksheetstream'] = add_exercises(ctx['worksheetstream'], ctx, req)
+
+class SubjectMediaView(BaseView):
+    '''The view of subject media files.
+
+    URIs pointing here will just be served directly, from the subject's
+    media directory.
+    '''
+
+    def __init__(self, req, subject, path):
+        self.subject = req.store.find(Subject, code=subject).one()
+        self.path = os.path.normpath(path)
+
+    def render(self, req):
+        # If the subject doesn't exist, self.subject will be None. Die.
+        if not self.subject:
+            raise NotFound()
+
+        # If it begins with ".." or separator, it's illegal. Die.
+        if self.path.startswith("..") or self.path.startswith('/'):
+            raise Forbidden()
+        subjectdir = os.path.join(ivle.conf.subjects_base,
+                                  self.subject.code, 'media')
+        filename = os.path.join(subjectdir, self.path)
+
+        # Find an appropriate MIME type.
+        (type, _) = mimetypes.guess_type(filename)
+        if type is None:
+            type = ivle.conf.mimetypes.default_mimetype
+
+        # Get out if it is unreadable or a directory.
+        if not os.access(filename, os.F_OK):
+            raise NotFound()
+        if not os.access(filename, os.R_OK) or os.path.isdir(filename):
+            raise Forbidden()
+
+        req.content_type = type
+        req.sendfile(filename)
 
 def is_valid_subjname(subject):
     m = re_ident.match(subject)
     return m is not None and m.end() == len(subject)
 
-def handle_subject_menu(req, subject):
-    # This is represented as a directory. Redirect and add a slash if it is
-    # missing.
-    if req.uri[-1] != '/':
-        req.throw_redirect(make_tutorial_path(subject))
-    # Subject names must be valid identifiers
-    if not is_valid_subjname(subject):
-        req.throw_error(req.HTTP_NOT_FOUND,
-            "Invalid subject name: %s." % repr(subject))
-    # Parse the subject description file
-    # The subject directory must have a file "subject.xml" in it,
-    # or it does not exist (404 error).
-    try:
-        subjectfile = open(os.path.join(ivle.conf.subjects_base, subject,
-            "subject.xml"))
-    except:
-        req.throw_error(req.HTTP_NOT_FOUND,
-            "Subject %s not found." % repr(subject))
-
-    # Read in data about the subject
-    subjectdom = minidom.parse(subjectfile)
-    subjectfile.close()
-    # TEMP: All of this is for a temporary XML format, which will later
-    # change.
-    worksheetsdom = subjectdom.documentElement
-    worksheets = []     # List of string IDs
-    for worksheetdom in worksheetsdom.childNodes:
-        if worksheetdom.nodeType == worksheetdom.ELEMENT_NODE:
-            # Get the 3 attributes for this node and construct a Worksheet
-            # object.
-            # (Note: assessable will default to False, unless it is explicitly
-            # set to "true").
-            worksheet = Worksheet(worksheetdom.getAttribute("id"),
-                worksheetdom.getAttribute("name"),
-                worksheetdom.getAttribute("assessable") == "true")
-            worksheets.append(worksheet)
-
-    # Now all the errors are out the way, we can begin writing
-    req.title = "Tutorial - %s" % subject
-    req.write_html_head_foot = True
-    req.write('<div id="ivle_padding">\n')
-    req.write("<h1>IVLE Tutorials - %s</h1>\n" % cgi.escape(subject))
-    req.write('<h2>Worksheets</h2>\n<ul id="tutorial-toc">\n')
-    # As we go, calculate the total score for this subject
-    # (Assessable worksheets only, mandatory problems only)
-    problems_done = 0
-    problems_total = 0
-    for worksheet_from_xml in worksheets:
-        worksheet = ivle.database.Worksheet.get_by_name(req.store,
-            subject, worksheet_from_xml.id)
-        # If worksheet is not in database yet, we'll simply not display
-        # data about it yet (it should be added as soon as anyone visits
-        # the worksheet itself).
-        req.write('  <li><a href="%s">%s</a>'
-            % (urllib.quote(worksheet_from_xml.id),
-                cgi.escape(worksheet_from_xml.name)))
-        if worksheet is not None:
-            # If the assessable status of this worksheet has changed,
-            # update the DB
-            # (Note: This fails the try block if the worksheet is not yet
-            # in the DB, which is fine. The author should visit the
-            # worksheet page to get it into the DB).
-            if worksheet.assessable != worksheet_from_xml.assessable:
-                # XXX If statement to avoid unnecessary database writes.
-                # Is this necessary, or will Storm check for us?
-                worksheet.assessable = worksheet_from_xml.assessable
-                req.store.commit()
-            if worksheet.assessable:
-                # Calculate the user's score for this worksheet
-                mand_done, mand_total, opt_done, opt_total = (
-                    ivle.worksheet.calculate_score(req.store, req.user,
-                        worksheet))
-                if opt_total > 0:
-                    optional_message = " (excluding optional exercises)"
-                else:
-                    optional_message = ""
-                if mand_done >= mand_total:
-                    complete_class = "complete"
-                elif mand_done > 0:
-                    complete_class = "semicomplete"
-                else:
-                    complete_class = "incomplete"
-                problems_done += mand_done
-                problems_total += mand_total
-                req.write('\n    <ul><li class="%s">'
-                        'Completed %d/%d%s</li></ul>\n  '
-                        % (complete_class, mand_done, mand_total,
-                            optional_message))
-        req.write('</li>\n')
-    req.write("</ul>\n")
-    if problems_total > 0:
-        if problems_done >= problems_total:
-            complete_class = "complete"
-        elif problems_done > 0:
-            complete_class = "semicomplete"
-        else:
-            complete_class = "incomplete"
-        problems_pct = (100 * problems_done) / problems_total       # int
-        req.write('<ul><li class="%s">Total exercises completed: %d/%d '
-                    '(%d%%)</li></ul>\n'
-            % (complete_class, problems_done, problems_total,
-                problems_pct))
-        # XXX Marks calculation (should be abstracted out of here!)
-        # percent / 16, rounded down, with a maximum mark of 5
-        max_mark = 5
-        mark = min(problems_pct / 16, max_mark)
-        req.write('<p style="font-weight: bold">Worksheet mark: %d/%d'
-                    '</p>\n' % (mark, max_mark))
-    req.write("</div>\n")   # tutorialbody
-
-def handle_worksheet(req, subject, worksheet):
-    # Subject and worksheet names must be valid identifiers
-    if not is_valid_subjname(subject) or not is_valid_subjname(worksheet):
-        req.throw_error(req.HTTP_NOT_FOUND,
-            "Invalid subject name %s or worksheet name %s."
-                % (repr(subject), repr(worksheet)))
-
-    # Read in worksheet data
-    worksheetfilename = os.path.join(ivle.conf.subjects_base, subject,
-            worksheet + ".xml")
-    try:
-        worksheetfile = open(worksheetfilename)
-        worksheetmtime = os.path.getmtime(worksheetfilename)
-    except:
-        req.throw_error(req.HTTP_NOT_FOUND,
-            "Worksheet file not found.")
-    worksheetmtime = datetime.fromtimestamp(worksheetmtime)
-
-    worksheetdom = minidom.parse(worksheetfile)
-    worksheetfile.close()
-    # TEMP: All of this is for a temporary XML format, which will later
-    # change.
-    worksheetdom = worksheetdom.documentElement
-    if worksheetdom.tagName != "worksheet":
-        req.throw_error(req.HTTP_INTERNAL_SERVER_ERROR,
-            "The worksheet XML file's top-level element must be <worksheet>.")
-    worksheetname = worksheetdom.getAttribute("name")
-
-    # Now all the errors are out the way, we can begin writing
-    req.title = "Tutorial - %s" % worksheetname
-    req.write_html_head_foot = True
-    req.write('<div id="ivle_padding">\n')
-    req.write("<h1>IVLE Tutorials - %s</h1>\n<h2>%s</h2>\n"
-        % (cgi.escape(subject), cgi.escape(worksheetname)))
-    exercise_list = present_table_of_contents(req, worksheetdom, 0)
-    # If the database is missing this worksheet or out of date, update its
-    # details about this worksheet
-    # Note: Do NOT set assessable (this is done at the subject level).
-    update_db_worksheet(req.store, subject, worksheet, worksheetmtime,
-        exercise_list)
-
-    # Write each element
-    exerciseid = 0
-    for node in worksheetdom.childNodes:
-        exerciseid = present_worksheet_node(req, node, exerciseid)
-    req.write("</div>\n")   # tutorialbody
-
-def present_table_of_contents(req, node, exerciseid):
-    """Given a node of a worksheet XML document, writes out a table of
-    contents to the request. This recursively searches for "excercise"
-    and heading elements to write out.
-
-    When exercise elements are encountered, the DB is queried for their
-    completion status, and the ball is shown of the appropriate colour.
-
-    exerciseid is the ID to use for the first exercise.
-
-    As a secondary feature, this records the identifier (xml filename) and
-    optionality of each exercise in a list of pairs [(str, bool)], and returns
-    this list. This can be used to cache this information in the database.
-    """
-    exercise_list = []
-    # XXX This means the DB is queried twice for each element.
-    # Consider caching these results for lookup later.
-    req.write("""<div id="tutorial-toc">
-<h2>Worksheet Contents</h2>
-<ul>
-""")
-    for tag, xml in find_all_nodes(req, node):
-        if tag == "ex":
-            # Exercise node
-            # Fragment ID is an accumulating exerciseid
-            # (The same algorithm is employed when presenting exercises)
-            fragment_id = "exercise%d" % exerciseid
-            exerciseid += 1
-            exercisesrc = xml.getAttribute("src")
-            # Optionality: Defaults to False
-            exerciseoptional = xml.getAttribute("optional") == "true"
-            # Record the name and optionality for returning in the list
-            exercise_list.append((exercisesrc, exerciseoptional))
-            # TODO: Get proper exercise title
-            title = exercisesrc
-            # Get the completion status of this exercise
-            exercise = ivle.database.Exercise.get_by_name(req.store,
-                            exercisesrc)
-            complete, _ = ivle.worksheet.get_exercise_status(req.store,
-                            req.user, exercise)
-            req.write('  <li class="%s" id="toc_li_%s"><a href="#%s">%s'
-                '</a></li>\n'
-                % ("complete" if complete else "incomplete",
-                    fragment_id, fragment_id, cgi.escape(title)))
-        else:
-            # Heading node
-            fragment_id = getID(xml)
-            title = getTextData(xml)
-            req.write('  <li><a href="#%s">%s</a></li>\n'
-                % (fragment_id, cgi.escape(title)))
-    req.write('</ul>\n</div>\n')
-    return exercise_list
-
-def find_all_nodes(req, node):
-    """Generator. Searches through a node and yields all headings and
-    exercises. (Recursive).
-    When finding a heading, yields a pair ("hx", headingnode), where "hx" is
-    the element name, such as "h1", "h2", etc.
-    When finding an exercise, yields a pair ("ex", exercisenode), where
-    exercisenode is the DOM node for this exercise.
-    """
-    if node.nodeType == node.ELEMENT_NODE:
-        if node.tagName == "exercise":
-            yield "ex", node
-        elif (node.tagName == "h1" or node.tagName == "h2"
-            or node.tagName == "h3"):
-            yield node.tagName, node
-        else:
-            # Some other element. Recurse.
-            for childnode in node.childNodes:
-                for yieldval in find_all_nodes(req, childnode):
-                    yield yieldval
-
-def present_worksheet_node(req, node, exerciseid):
-    """Given a node of a worksheet XML document, writes it out to the
-    request. This recursively searches for "exercise" elements and handles
-    those specially (presenting their XML exercise spec and input box), and
-    just dumps the other elements as regular HTML.
-
-    exerciseid is the ID to use for the first exercise.
-    Returns the new exerciseid after all the exercises have been written
-    (since we need unique IDs for each exercise).
-    """
-    if node.nodeType == node.ELEMENT_NODE:
-        if node.tagName == "exercise":
-            present_exercise(req, node.getAttribute("src"), exerciseid)
-            exerciseid += 1
-        else:
-            # Some other element. Write out its head and foot, and recurse.
-            req.write("<" + node.tagName)
-            # Attributes
-            attrs = map(lambda (k,v): '%s="%s"'
-                    % (cgi.escape(k), cgi.escape(v)), node.attributes.items())
-            if len(attrs) > 0:
-                req.write(" " + ' '.join(attrs))
-            req.write(">")
-            for childnode in node.childNodes:
-                exerciseid = present_worksheet_node(req, childnode, exerciseid)
-            req.write("</" + node.tagName + ">")
-    else:
-        # No need to recurse, so just print this node's contents
-        req.write(node.toxml())
-    return exerciseid
+def get_worksheets(subjectfile):
+    '''Given a subject stream, get all the worksheets and put them in ctx'''
+    worksheets = []
+    for kind, data, pos in subjectfile:
+        if kind is genshi.core.START:
+            if data[0] == 'worksheet':
+                worksheetid = ''
+                worksheetname = ''
+                worksheetasses = False
+                for attr in data[1]:
+                    if attr[0] == 'id':
+                        worksheetid = attr[1]
+                    elif attr[0] == 'name':
+                        worksheetname = attr[1]
+                    elif attr[0] == 'assessable':
+                        worksheetasses = attr[1] == 'true'
+                worksheets.append(Worksheet(worksheetid, worksheetname, \
+                                                            worksheetasses))
+    return worksheets
+
+# This generator adds in the exercises as they are required. This is returned    
+def add_exercises(stream, ctx, req):
+    """A filter adds exercises into the stream."""
+    exid = 0
+    for kind, data, pos in stream:
+        if kind is genshi.core.START:
+            if data[0] == 'exercise':
+                new_stream = ctx['exercises'][exid]['stream']
+                exid += 1
+                for item in new_stream:
+                    yield item
+            else:
+                yield kind, data, pos
+        else:
+            yield kind, data, pos
+
+# This function runs through the worksheet, to get data on the exercises to
+# build a Table of Contents, as well as fill in details in ctx
+def generate_worksheet_data(ctx, req):
+    """Runs through the worksheetstream, generating the exericises"""
+    exid = 0
+    ctx['exercises'] = []
+    ctx['exerciselist'] = []
+    for kind, data, pos in ctx['worksheetstream']:
+        if kind is genshi.core.START:
+            if data[0] == 'exercise':
+                exid += 1
+                src = ""
+                optional = False
+                for attr in data[1]:
+                    if attr[0] == 'src':
+                        src = attr[1]
+                    if attr[0] == 'optional':
+                        optional = attr[1] == 'true'
+                # Each item in toc is of type (name, complete, stream)
+                ctx['exercises'].append(present_exercise(req, src, exid))
+                ctx['exerciselist'].append((src, optional))
+            elif data[0] == 'worksheet':
+                ctx['worksheetname'] = 'bob'
+                for attr in data[1]:
+                    if attr[0] == 'name':
+                        ctx['worksheetname'] = attr[1]
 
 def innerXML(elem):
     """Given an element, returns its children as XML strings concatenated
@@ -448,21 +329,6 @@
         s += child.toxml()
     return s
 
-def getID(element):
-    """Get the first ID attribute found when traversing a node and its
-    children. (This is used to make fragment links to a particular element).
-    Returns None if no ID is found.
-    """
-    id = element.getAttribute("id")
-    if id is not None and id != '':
-        return id
-    for child in element.childNodes:
-        if child.nodeType == child.ELEMENT_NODE:
-            id = getID(child)
-            if id is not None:
-                return id
-    return None
-
 def getTextData(element):
     """ Get the text and cdata inside an element
     Leading and trailing whitespace are stripped
@@ -478,134 +344,79 @@
 
     return data.strip()
 
+#TODO: This needs to be re-written, to stop using minidom, and get the data
+# about the worksheet directly from the database
 def present_exercise(req, exercisesrc, exerciseid):
     """Open a exercise file, and write out the exercise to the request in HTML.
     exercisesrc: "src" of the exercise file. A path relative to the top-level
         exercises base directory, as configured in conf.
     """
-    req.write('<div class="exercise" id="exercise%d">\n'
-        % exerciseid)
+    # Exercise-specific context is used here, as we already have all the data
+    # we need
+    curctx = genshi.template.Context()
+    curctx['filename'] = exercisesrc
+    curctx['exerciseid'] = exerciseid
+
+    # Retrieve the exercise details from the database
     exercise = ivle.database.Exercise.get_by_name(req.store, exercisesrc)
+    #Open the exercise, and double-check that it exists
     exercisefile = util.open_exercise_file(exercisesrc)
     if exercisefile is None:
-        req.write("<p><b>Server Error</b>: "
-            + "Exercise file could not be opened.</p>\n")
-        req.write("</div>\n")
-        return
+        req.throw_error(req.HTTP_EXPECTATION_FAILED, \
+                                        "Exercise file could not be opened")
     
     # Read exercise file and present the exercise
     # Note: We do not use the testing framework because it does a lot more
     # work than we need. We just need to get the exercise name and a few other
     # fields from the XML.
 
+    #TODO: Replace calls to minidom with calls to the database directly
     exercisedom = minidom.parse(exercisefile)
     exercisefile.close()
     exercisedom = exercisedom.documentElement
     if exercisedom.tagName != "exercise":
         req.throw_error(req.HTTP_INTERNAL_SERVER_ERROR,
             "The exercise XML file's top-level element must be <exercise>.")
-    exercisename = exercisedom.getAttribute("name")
-    rows = exercisedom.getAttribute("rows")
-    if not rows:
-        rows = "12"
+    curctx['exercisename'] = exercisedom.getAttribute("name")
+    
+    curctx['rows'] = exercisedom.getAttribute("rows")
+    if not curctx['rows']:
+        curctx['rows'] = "12"
     # Look for some other fields we need, which are elements:
     # - desc
     # - partial
-    exercisedesc = None
-    exercisepartial= ""
+    curctx['exercisedesc'] = None
+    curctx['exercisepartial'] = ""
     for elem in exercisedom.childNodes:
         if elem.nodeType == elem.ELEMENT_NODE:
             if elem.tagName == "desc":
-                exercisedesc = rst(innerXML(elem).strip())
+                curctx['exercisedesc'] = genshi.XML(rst(innerXML(elem).strip()))
             if elem.tagName == "partial":
-                exercisepartial= getTextData(elem) + '\n'
-    exercisepartial_backup = exercisepartial
+                curctx['exercisepartial'] = getTextData(elem) + '\n'
+    curctx['exercisepartial_backup'] = curctx['exercisepartial']
 
     # If the user has already saved some text for this problem, or submitted
     # an attempt, then use that text instead of the supplied "partial".
     saved_text = ivle.worksheet.get_exercise_stored_text(req.store,
         req.user, exercise)
     # Also get the number of attempts taken and whether this is complete.
-    complete, attempts = ivle.worksheet.get_exercise_status(req.store,
-        req.user, exercise)
+    complete, curctx['attempts'] = \
+            ivle.worksheet.get_exercise_status(req.store, req.user, exercise)
     if saved_text is not None:
-        exercisepartial = saved_text.text
-
-    # Print this exercise out to HTML 
-    req.write("<p><b>Exercise:</b> %s</p>\n" % cgi.escape(exercisename))
-    if exercisedesc is not None:
-        req.write("<div>%s</div>\n" % exercisedesc)
-    filename = cgi.escape(cjson.encode(exercisesrc), quote=True)
-    req.write("""<input id="input_resettext_exercise%d" type="hidden"
-    value="%s" />"""
-        % (exerciseid, urllib.quote(exercisepartial_backup)))
-    req.write("""<textarea id="textarea_exercise%d" class="exercisebox"
-    onkeypress="return catch_textbox_input(&quot;exercise%d&quot;, %s,
-        event.keyCode)"
-    onchange="set_saved_status(&quot;exercise%d&quot;, %s,
-        &quot;Save&quot;)"
-    cols="80" rows="%s">%s</textarea>"""
-        % (exerciseid, exerciseid, filename, exerciseid, filename,
-            rows, cgi.escape(exercisepartial)))
-    req.write("""\n<div class="exercisebuttons">\n""")
-    req.write("""  <input type="button" value="Saved" disabled="disabled"
-    id="savebutton_exercise%d"
-    onclick="saveexercise(&quot;exercise%d&quot;, %s)"
-    title="Save your solution to this exercise" />\n"""
-        % (exerciseid, exerciseid, filename))
-    req.write("""  <input type="button" value="Reset"
-    id="resetbutton_exercise%d"
-    onclick="resetexercise(&quot;exercise%d&quot;, %s)"
-    title="Reload the original partial solution for this exercise" />\n"""
-        % (exerciseid, exerciseid, filename))
-    req.write("""  <input type="button" value="Run"
-    onclick="runexercise(&quot;exercise%d&quot;, %s)"
-    title="Run this program in the console" />\n"""
-        % (exerciseid, filename))
-    req.write("""  <input type="button" value="Submit"
-    id="submitbutton_exercise%d"
-    onclick="submitexercise(&quot;exercise%d&quot;, %s)"
-    title="Submit this solution for evaluation" />\n"""
-        % (exerciseid, exerciseid, filename))
-    req.write("""</div>
-<div class="testoutput">
-</div>
-""")
-    # Write the "summary" - whether this problem is complete and how many
-    # attempts it has taken.
-    req.write("""<div class="problem_summary">
-  <ul><li id="summaryli_exercise%d" class="%s">
-    <b><span id="summarycomplete_exercise%d">%s</span>.</b>
-    Attempts: <span id="summaryattempts_exercise%d">%d</span>.
-  </li></ul>
-</div>
-""" % (exerciseid, "complete" if complete else "incomplete",
-        exerciseid, "Complete" if complete else "Incomplete",
-        exerciseid, attempts))
-    # Write the attempt history infrastructure
-    req.write("""<div class="attempthistory">
-  <p><a title="Click to view previous submissions you have made for this \
-exercise" onclick="open_previous(&quot;exercise%d&quot;, %s)">View previous \
-attempts</a></p>
-  <div style="display: none">
-    <h3>Previous attempts</h3>
-    <p><a title="Close the previous attempts" \
-onclick="close_previous(&quot;exercise%d&quot;)">Close attempts</a></p>
-    <p>
-      <select title="Select an attempt's time stamp from the list">
-        <option></option>
-      </select>
-      <input type="button" value="View"
-        onclick="select_attempt(&quot;exercise%d&quot;, %s)" />
-    </p>
-    <p><textarea readonly="readonly" class="exercisebox" cols="80" rows="%s"
-        title="You submitted this code on a previous attempt">
-       </textarea>
-    </p>
-  </div>
-</div>
-""" % (exerciseid, filename, exerciseid, exerciseid, filename, rows))
-    req.write("</div>\n")
+        curctx['exercisepartial'] = saved_text.text
+    if complete:
+        curctx['complete'] = 'complete'
+    else:
+        curctx['complete'] = 'incomplete'
+
+    #Save the exercise details to the Table of Contents
+
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(os.path.join(os.path.dirname(__file__), "exercise.html"))
+    ex_stream = tmpl.generate(curctx)
+    return {'name': curctx['exercisename'], 'complete': curctx['complete'], \
+              'stream': ex_stream, 'exid': exerciseid}
+
 
 def update_db_worksheet(store, subject, worksheetname, file_mtime,
     exercise_list=None, assessable=None):
@@ -657,3 +468,10 @@
                     worksheet=worksheet, exercise=exercise, optional=optional)
 
     store.commit()
+
+class Plugin(BasePlugin):
+    urls = [
+        ('subjects/:subject/+worksheets', SubjectView),
+        ('subjects/:subject/+worksheets/+media/*(path)', SubjectMediaView),
+        ('subjects/:subject/+worksheets/:worksheet', WorksheetView),
+    ]

=== added file 'ivle/webapp/tutorial/exercise.html'
--- ivle/webapp/tutorial/exercise.html	1970-01-01 00:00:00 +0000
+++ ivle/webapp/tutorial/exercise.html	2009-02-04 13:01:16 +0000
@@ -0,0 +1,70 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <div class="exercise" id="exercise${exerciseid}">
+    <p><b>Exercise:</b> ${exercisename}</p>
+    <py:if test="exercisedesc is not None">
+      <div>${exercisedesc}</div>
+    </py:if>
+    <input id="input_resettext_exercise${exerciseid}" type="hidden" value="${exercisepartial_backup}" />
+    <textarea id="textarea_exercise${exerciseid}" class="exercisebox"
+        onkeypress="return catch_textbox_input(&quot;exercise${exerciseid}&quot;, '${filename}',
+        event.keyCode)"
+        onchange="set_saved_status(&quot;exercise${exerciseid}&quot;, '${filename}',
+        &quot;Save&quot;)"
+        cols="80" rows="${rows}">${exercisepartial}</textarea>
+    <div class="exercisebuttons">
+      <input type="button" value="Saved" disabled="disabled" id="savebutton_exercise${exerciseid}"
+          onclick="saveexercise(&quot;exercise${exerciseid}&quot;, '${filename}')"
+          title="Save your solution to this exercise" />
+      <input type="button" value="Reset" id="resetbutton_exercise${exerciseid}"
+          onclick="resetexercise(&quot;exercise${exerciseid}&quot;, '${filename}')"
+          title="Reload the original partial solution for this exercise" />
+      <input type="button" value="Run"
+          onclick="runexercise(&quot;exercise${exerciseid}&quot;, '${filename}')"
+          title="Run this program in the console" />
+      <input type="button" value="Submit"
+              id="submitbutton_exercise${exerciseid}"
+              onclick="submitexercise(&quot;exercise${exerciseid}&quot;, '${filename}')"
+              title="Submit this solution for evaluation" />
+    </div>
+    <div class="testoutput">
+    </div>
+    <div class="problem_summary">
+      <ul>
+        <li id="summaryli_exercise${exerciseid}" class="${complete}">
+          <b><span id="summarycomplete_exercise${exerciseid}">${complete}</span>.</b>
+            Attempts: <span id="summaryattempts_exercise${exerciseid}">${attempts}</span>.
+        </li>
+      </ul>
+    </div>
+    <!-- This is filled in by javascript when the page loads -->
+    <div class="attempthistory">
+      <p>
+        <a title="Click to view previous submissions you have made for this exercise" 
+              onclick="open_previous(&quot;exercise${exerciseid}&quot;, '${filename}')">
+          View previous attempts
+        </a>
+      </p>
+    <div style="display: none">
+      <h3>Previous attempts</h3>
+        <p>
+          <a title="Close the previous attempts" onclick="close_previous(&quot;exercise${exerciseid}&quot;)">
+            Close attempts
+          </a>
+        </p>
+        <p>
+          <select title="Select an attempt's time stamp from the list">
+            <option></option>
+          </select>
+          <input type="button" value="View"
+              onclick="select_attempt(&quot;exercise${exerciseid}&quot;, '${filename}')" />
+        </p>
+        <p>
+          <textarea readonly="readonly" class="exercisebox" cols="80" rows="${rows}"
+              title="You submitted this code on a previous attempt">
+          </textarea>
+        </p>
+      </div>
+    </div>
+  </div>
+</html>

=== added file 'ivle/webapp/tutorial/subjectmenu.html'
--- ivle/webapp/tutorial/subjectmenu.html	1970-01-01 00:00:00 +0000
+++ ivle/webapp/tutorial/subjectmenu.html	2009-02-04 06:53:03 +0000
@@ -0,0 +1,38 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <head>
+    <title>Tutorial - Subject Menu</title>
+  </head>
+  <body>
+    <div id="ivle_padding">
+      <h1>IVLE Tutorials - ${subject}</h1>
+      <h2>Worksheets</h2>
+      <ul id="tutorial-toc">
+        <py:for each="worksheet in worksheets">
+          <li>
+            <!-- XXX: Generate URL properly! -->
+            <a href="/subjects/${subject}/+worksheets/${worksheet.id}">${worksheet.name}</a>
+            <py:if test="worksheet.assessable">
+              <ul>
+                <li class="${worksheet.complete_class}">
+                  Completed ${worksheet.mand_done}/${worksheet.total}${worksheet.optional_message}
+                </li>
+              </ul>
+            </py:if>
+          </li>
+        </py:for>
+      </ul>
+      <py:if test="problems_total > 0">
+        <ul>
+          <li class="${complete_class}">
+            Total exercises completed: ${problems_done}/${problems_total} (${problems_pct}%)
+          </li>
+        </ul> 
+        <p style="font-weight: bold">
+          Worksheet mark: ${mark}/${max_mark}
+        </p>
+      </py:if>
+    </div>            
+  </body>
+</html>

=== added file 'ivle/webapp/tutorial/worksheet.html'
--- ivle/webapp/tutorial/worksheet.html	1970-01-01 00:00:00 +0000
+++ ivle/webapp/tutorial/worksheet.html	2009-02-04 06:53:03 +0000
@@ -0,0 +1,28 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <head>
+    <title>Tutorial - Worksheet</title>
+  </head>
+  <body>
+    <div id="ivle_padding">
+      <h1>IVLE Tutorials - ${subject}</h1>
+      <!-- Display the Table Of Contents -->
+      <div id="tutorial-toc">
+        <h2>Worksheet Contents</h2>
+        <ul>
+          <py:for each="exercise in exercises">
+            <li class="${exercise['complete']}" id="toc_li_exercise${exercise['exid']}">
+              <a href="#exercise${exercise['exid']}">
+                ${exercise.name}
+              </a>
+            </li>
+          </py:for>
+        </ul>
+      </div>
+      <!-- Display The Worksheet -->
+      <h2>${worksheetname}</h2>
+      ${worksheetstream}
+    </div>
+  </body>
+</html>

=== modified file 'setup.py'
--- setup.py	2009-02-02 05:14:04 +0000
+++ setup.py	2009-01-13 01:36:15 +0000
@@ -31,6 +31,7 @@
 import setup.build
 import setup.install
 
+
 def main(argv=None):
     if argv is None:
         argv = sys.argv
@@ -80,6 +81,7 @@
             'config' : setup.configure.configure,
             'build' : setup.build.build,
             'install' : setup.install.install,
+            #'updatejails' : None,
         }[operation]
     except KeyError:
         print >>sys.stderr, (

=== modified file 'setup/build.py'
--- setup/build.py	2009-02-02 05:14:04 +0000
+++ setup/build.py	2009-01-13 01:36:15 +0000
@@ -29,32 +29,50 @@
 def build(args):
     usage = """usage: %prog build [options]
 (requires root)
-Compiles platform-specific code, and optionally Python too.
+Compiles all files and sets up a jail template in the source directory.
+-O is recommended to cause compilation to be optimised.
 Details:
 Compiles (GCC) bin/trampoline/trampoline.c to bin/trampoline/trampoline.
 Compiles (GCC) bin/timount/timount.c to bin/timount/timount.
-Optionally generates .pyc files for all the IVLE .py files."""
+Creates jail with system and student packages installed from MIRROR.
+Copies console/ to a location within the jail.
+Copies OS programs and files to corresponding locations within the jail
+  (eg. python and Python libs, ld.so, etc).
+Generates .pyc or .pyo files for all the IVLE .py files."""
 
     # Parse arguments
     parser = optparse.OptionParser(usage)
     parser.add_option("-n", "--dry",
         action="store_true", dest="dry",
         help="Print out the actions but don't do anything.")
-    parser.add_option("--no-compile",
-        action="store_true", dest="nocompile",
-        help="Don't byte-compile .py files.")
+    parser.add_option("-j", "--rebuildjail",
+        action="store_true", dest="rebuildjail",
+        help="Don't recreate jail/ - just update its IVLE code.")
+    parser.add_option("-m", "--mirror",
+        action="store", dest="apt_mirror",
+        help="Sets the APT mirror used to build the jail.")
     (options, args) = parser.parse_args(args)
 
     # Call the real function
-    return __build(options.dry, options.nocompile)
+    return __build(options.dry, options.rebuildjail, options.apt_mirror)
 
-def __build(dry=False, no_compile=None):
+def __build(dry=False,rebuildjail=False,apt_mirror=None):
     # We need to import the one in the working copy, not in the system path.
     confmodule = __import__("ivle/conf/conf")
     install_list = util.InstallList()
 
+    # Must be run as root or a dry run  
     if dry:
         print "Dry run (no actions will be executed)\n"
+    
+    if not dry and os.geteuid() != 0:
+        print >>sys.stderr, "Must be root to run build"
+        print >>sys.stderr, "(I need to chroot)."
+        return 1
+    
+    if not rebuildjail and not os.path.exists('jail'):
+        print >> sys.stderr, "No jail exists -- please rerun with -j."
+        return 1
 
     # Compile the trampoline
     curdir = os.getcwd()
@@ -68,11 +86,41 @@
     util.action_runprog('make', [], dry)
     os.chdir(curdir)
 
+    if rebuildjail:
+        # Create the jail and its subdirectories
+        # Note: Other subdirs will be made by copying files
+        if apt_mirror != None:
+            os.environ['MIRROR'] = apt_mirror
+        util.action_runprog('setup/buildjail.sh', [], dry)
+
+    # Copy all console and operating system files into the jail
+    jail_share = os.path.join('jail', confmodule.share_path[1:])
+    jail_services = os.path.join(jail_share, 'services')
+    util.action_copylist(install_list.list_services, jail_share, dry)
+
+    # Chmod the python console
+    util.action_chmod_x(os.path.join(jail_services, 'python-console'), dry)
+    util.action_chmod_x(os.path.join(jail_services, 'fileservice'), dry)
+    util.action_chmod_x(os.path.join(jail_services, 'serveservice'), dry)
+
+    # Also copy the IVLE lib directory into the jail
+    # This is necessary for running certain services
+    jail_site_packages = os.path.join('jail',
+                                      confmodule.python_site_packages[1:])
+    util.action_copylist(install_list.list_ivle_lib, jail_site_packages, dry)
+    # IMPORTANT: ivle/conf/conf.py contains details
+    # which could compromise security if left in the jail (such as the DB
+    # password).
+    # The "safe" version is in jailconf.py. Delete conf.py and replace it with
+    # jailconf.py.
+    util.action_copyfile('ivle/conf/jailconf.py',
+        os.path.join(jail_site_packages, 'ivle/conf/conf.py'), dry)
+
     # Compile .py files into .pyc or .pyo files
-    if not no_compile:
-        compileall.compile_dir('www', quiet=True)
-        compileall.compile_dir('ivle', quiet=True)
-        compileall.compile_dir('services', quiet=True)
+    compileall.compile_dir('www', quiet=True)
+    compileall.compile_dir('ivle', quiet=True)
+    compileall.compile_dir('services', quiet=True)
+    compileall.compile_dir(os.path.join(jail_site_packages, 'ivle'),quiet=True)
 
     return 0
 

=== modified file 'setup/configure.py'
--- setup/configure.py	2009-02-02 05:14:04 +0000
+++ setup/configure.py	2009-01-15 01:29:07 +0000
@@ -1,5 +1,5 @@
 # IVLE - Informatics Virtual Learning Environment
-# Copyright (C) 2007-2009 The University of Melbourne
+# Copyright (C) 2007-2008 The University of Melbourne
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -19,12 +19,11 @@
 # Author: Matt Giuca, Refactored by David Coles
 # Date:   03/07/2008
 
-'''Configures IVLE with machine-specific details, most notably, various paths.
-Either prompts the administrator for these details or accepts them as
-command-line args.
-
-Creates ivle/conf/conf.py and bin/trampoline/trampoline/conf.h.
-'''
+# setup/config.py
+# Configures IVLE with machine-specific details, most notably, various paths.
+# Either prompts the administrator for these details or accepts them as
+# command-line args.
+# Creates lib/conf/conf.py and bin/trampoline/conf.h.
 
 import optparse
 import getopt
@@ -221,6 +220,14 @@
 # The password for the usrmgt-server.""", ask=False))
 
 def configure(args):
+    usage = """usage: %prog config [options]
+Creates lib/conf/conf.py (and a few other config files).
+Interactively asks questions to set this up."""
+
+    # Parse arguments
+    parser = optparse.OptionParser(usage)
+    (options, args) = parser.parse_args(args)
+
     # Call the real function
     return __configure(args)
 
@@ -249,6 +256,7 @@
 
     # the files that will be created/overwritten
     conffile = os.path.join(cwd, "ivle/conf/conf.py")
+    jailconffile = os.path.join(cwd, "ivle/conf/jailconf.py")
     conf_hfile = os.path.join(cwd, "bin/trampoline/conf.h")
     phpBBconffile = os.path.join(cwd, "www/php/phpBB3/config.php")
 
@@ -270,11 +278,12 @@
     %s
     %s
     %s
+    %s
 prompting you for details about your configuration. The file will be
 overwritten if it already exists. It will *not* install or deploy IVLE.
 
 Please hit Ctrl+C now if you do not wish to do this.
-""" % (conffile, conf_hfile, phpBBconffile)
+""" % (conffile, jailconffile, conf_hfile, phpBBconffile)
 
         # Get information from the administrator
         # If EOF is encountered at any time during the questioning, just exit
@@ -351,6 +360,46 @@
 
     print "Successfully wrote %s" % conffile
 
+    # Write conf/jailconf.py
+
+    try:
+        conf = open(jailconffile, "w")
+
+        # In the "in-jail" version of conf, we don't need MOST of the details
+        # (it would be a security risk to have them here).
+        # So we just write root_dir, and jail_base is "/".
+        # (jail_base being "/" means "jail-relative" paths are relative to "/"
+        # when inside the jail.)
+        conf.write("""# IVLE Configuration File
+# conf.py
+# Miscellaneous application settings
+# (User jail version)
+
+
+# In URL space, where in the site is IVLE located. (All URLs will be prefixed
+# with this).
+# eg. "/" or "/ivle".
+root_dir = %s
+
+# In the local file system, where are the student/user file spaces located.
+# The user jails are expected to be located immediately in subdirectories of
+# this location.
+jail_base = '/'
+
+# The hostname for serving publicly accessible pages
+public_host = %s
+
+# The URL under which the Subversion repositories are located.
+svn_addr = %s
+""" % (repr(root_dir),repr(public_host),repr(svn_addr)))
+
+        conf.close()
+    except IOError, (errno, strerror):
+        print "IO error(%s): %s" % (errno, strerror)
+        sys.exit(1)
+
+    print "Successfully wrote %s" % jailconffile
+
     # Write bin/trampoline/conf.h
 
     try:
@@ -439,6 +488,7 @@
     print
     print "You may modify the configuration at any time by editing"
     print conffile
+    print jailconffile
     print conf_hfile
     print phpBBconffile
     print
@@ -482,9 +532,6 @@
 # In the local file system, where the template system jail will be stored.
 jail_system = os.path.join(jail_src_base, '__base__')
 
-# In the local file system, where the template system jail will be stored.
-jail_system_build = os.path.join(jail_src_base, '__base_build__')
-
 # In the local file system, where the subject content files are located.
 # (The 'subjects' and 'exercises' directories).
 content_path = os.path.join(data_path, 'content')

=== modified file 'setup/install.py'
--- setup/install.py	2009-02-02 05:14:04 +0000
+++ setup/install.py	2009-01-19 08:14:25 +0000
@@ -22,7 +22,6 @@
 import optparse
 import os
 import sys
-import functools
 
 from setup import util
 
@@ -35,6 +34,7 @@
 Copy bin/timount/timount to $target/bin.
 chown and chmod the installed trampoline.
 Copy www/ to $target.
+Copy jail/ to jail_system directory (unless --nojail specified).
 Copy subjects/ to subjects directory (unless --nosubjects specified).
 """
 
@@ -43,40 +43,31 @@
     parser.add_option("-n", "--dry",
         action="store_true", dest="dry",
         help="Print out the actions but don't do anything.")
+    parser.add_option("-J", "--nojail",
+        action="store_true", dest="nojail",
+        help="Don't copy jail/ to jail_system directory")
     parser.add_option("-S", "--nosubjects",
         action="store_true", dest="nosubjects",
         help="Don't copy subject/ to subjects directory.")
-    parser.add_option("-R", "--nosvnrevno",
-        action="store_true", dest="nosvnrevno",
-        help="Don't write out the Subversion revision to the share directory.")
-    parser.add_option("--root",
-        action="store", dest="rootdir",
-        help="Install into a different root directory.",
-        default='/')
     (options, args) = parser.parse_args(args)
 
     # Call the real function
-    return __install(options.dry, options.nosubjects, options.rootdir,
-                     options.nosvnrevno)
+    return __install(options.dry, options.nojail, options.nosubjects)
 
-def __install(dry=False, nosubjects=False, rootdir=None, nosvnrevno=False):
+def __install(dry=False,nojail=False,nosubjects=False):
     # We need to import the one in the working copy, not in the system path.
     confmodule = __import__("ivle/conf/conf")
     install_list = util.InstallList()
 
-    # We need to apply make_install_path with the rootdir to an awful lot of
-    # config variables, so make it easy:
-    mip = functools.partial(util.make_install_path, rootdir)
-
     # Pull the required varibles out of the config
-    lib_path = mip(confmodule.lib_path)
-    share_path = mip(confmodule.share_path)
-    bin_path = mip(confmodule.bin_path)
-    python_site_packages = mip(confmodule.python_site_packages)
-    jail_base = mip(confmodule.jail_base)
-    jail_system = mip(confmodule.jail_system)
-    subjects_base = mip(confmodule.subjects_base)
-    exercises_base = mip(confmodule.exercises_base)
+    lib_path = confmodule.lib_path
+    share_path = confmodule.share_path
+    bin_path = confmodule.bin_path
+    python_site_packages = confmodule.python_site_packages
+    jail_base = confmodule.jail_base
+    jail_system = confmodule.jail_system
+    subjects_base = confmodule.subjects_base
+    exercises_base = confmodule.exercises_base
 
     # Must be run as root or a dry run  
     if dry:
@@ -88,24 +79,24 @@
         return 1
 
     # Make some directories for data.
-    util.action_mkdir(mip(confmodule.log_path), dry)
-    util.action_mkdir(mip(confmodule.data_path), dry)
-    util.action_mkdir(mip(confmodule.jail_base), dry)
-    util.action_mkdir(mip(confmodule.jail_src_base), dry)
-    util.action_mkdir(mip(confmodule.content_path), dry)
-    util.action_mkdir(mip(confmodule.notices_path), dry)
-    util.action_mkdir(os.path.join(mip(confmodule.data_path), 'sessions'), dry)
-    util.action_mkdir(mip(confmodule.svn_path), dry)
-    util.action_mkdir(mip(confmodule.svn_repo_path), dry)
-    util.action_mkdir(os.path.join(mip(confmodule.svn_repo_path), 'users'),dry)
-    util.action_mkdir(os.path.join(mip(confmodule.svn_repo_path),'groups'),dry)
+    util.action_mkdir(confmodule.log_path, dry)
+    util.action_mkdir(confmodule.data_path, dry)
+    util.action_mkdir(confmodule.jail_base, dry)
+    util.action_mkdir(confmodule.jail_src_base, dry)
+    util.action_mkdir(confmodule.content_path, dry)
+    util.action_mkdir(confmodule.notices_path, dry)
+    util.action_mkdir(os.path.join(confmodule.data_path, 'sessions'), dry)
+    util.action_mkdir(confmodule.svn_path, dry)
+    util.action_mkdir(confmodule.svn_repo_path, dry)
+    util.action_mkdir(os.path.join(confmodule.svn_repo_path, 'users'), dry)
+    util.action_mkdir(os.path.join(confmodule.svn_repo_path, 'groups'), dry)
 
-    util.action_chown(mip(confmodule.log_path), util.wwwuid, util.wwwuid, dry)
-    util.action_chown(os.path.join(mip(confmodule.data_path), 'sessions'),
-                      util.wwwuid, util.wwwuid, dry)
-    util.action_chown(os.path.join(mip(confmodule.svn_repo_path), 'users'),
-                      util.wwwuid, util.wwwuid, dry)
-    util.action_chown(os.path.join(mip(confmodule.svn_repo_path), 'groups'),
+    util.action_chown(confmodule.log_path, util.wwwuid, util.wwwuid, dry)
+    util.action_chown(os.path.join(confmodule.data_path, 'sessions'),
+                      util.wwwuid, util.wwwuid, dry)
+    util.action_chown(os.path.join(confmodule.svn_repo_path, 'users'),
+                      util.wwwuid, util.wwwuid, dry)
+    util.action_chown(os.path.join(confmodule.svn_repo_path, 'groups'),
                       util.wwwuid, util.wwwuid, dry)
 
     # Create lib and copy the compiled files there
@@ -119,32 +110,39 @@
     timountpath = os.path.join(lib_path, 'timount')
     util.action_copyfile('bin/timount/timount', timountpath, dry)
 
-    # Copy in the services (only usrmgt-server is needed on the host, but
-    # the jail build requires the rest).
-    util.action_copylist(install_list.list_services, share_path, dry)
+    # Create a services directory to put the usrmgt-server in.
+    util.action_mkdir(os.path.join(share_path, 'services'), dry)
+
     usrmgtpath = os.path.join(share_path, 'services/usrmgt-server')
+    util.action_copyfile('services/usrmgt-server', usrmgtpath, dry)
     util.action_chmod_x(usrmgtpath, dry)
 
     # Copy the user-executable binaries using the list.
     util.action_copylist(install_list.list_user_binaries, bin_path, dry,
                          onlybasename=True)
 
-    # Copy the www directory (using the list)
+    # Copy the www and lib directories using the list
     util.action_copylist(install_list.list_www, share_path, dry)
+    util.action_copylist(install_list.list_ivle_lib, python_site_packages, dry)
+    
+    # Make the config file private
+    # XXX Get rid of lib
+    configpath = os.path.join(python_site_packages, 'ivle/conf/conf.py')
+    util.action_make_private(configpath, dry)
 
-    # Set appropriate permissions on the php directory
+    # Copy the php directory
     forum_dir = "www/php/phpBB3"
     forum_path = os.path.join(share_path, forum_dir)
+    util.action_copytree(forum_dir, forum_path, dry)
     print "chown -R www-data:www-data %s" % forum_path
     if not dry:
         os.system("chown -R www-data:www-data %s" % forum_path)
 
-    # Copy the lib directory (using the list)
-    util.action_copylist(install_list.list_ivle_lib, python_site_packages, dry)
-
-    # Make the config file private
-    configpath = os.path.join(python_site_packages, 'ivle/conf/conf.py')
-    util.action_make_private(configpath, dry)
+    if not nojail:
+        # Copy the local jail directory built by the build action
+        # to the jail_system directory (it will be used to help build
+        # all the students' jails).
+        util.action_copytree('jail', jail_system, dry)
 
     if not nosubjects:
         # Copy the subjects and exercises directories across
@@ -166,26 +164,25 @@
     except (IOError, OSError):
         pass
 
-    if not nosvnrevno:
-        # Create the ivle working revision record file
-        ivle_revision_file = os.path.join(share_path, 'revision.txt')
-        if not dry:
-            try:
-                conf = open(ivle_revision_file, "w")
+    # Create the ivle working revision record file
+    ivle_revision_record_file = os.path.join(share_path, 'ivle-revision.txt')
+    if not dry:
+        try:
+            conf = open(ivle_revision_record_file, "w")
 
-                conf.write("""# SVN revision r%s
+            conf.write("""# SVN revision r%s
 # Source tree location: %s
 # Modified files:
 """ % (util.get_svn_revision(), os.getcwd()))
 
-                conf.close()
-            except IOError, (errno, strerror):
-                print "IO error(%s): %s" % (errno, strerror)
-                sys.exit(1)
-
-            os.system("svn status . >> %s" % ivle_revision_file)
-
-        print "Wrote IVLE code revision status to %s" % ivle_revision_file
+            conf.close()
+        except IOError, (errno, strerror):
+            print "IO error(%s): %s" % (errno, strerror)
+            sys.exit(1)
+
+        os.system("svn status . >> %s" % ivle_revision_record_file)
+
+    print "Wrote IVLE code revision status to %s" % ivle_revision_record_file
 
     return 0
 

=== modified file 'setup/util.py'
--- setup/util.py	2009-02-02 05:14:04 +0000
+++ setup/util.py	2009-02-02 04:00:25 +0000
@@ -37,7 +37,7 @@
            'action_symlink', 'action_append', 'action_chown',
            'action_chown_setuid', 'action_chmod_x', 'action_make_private',
            'query_user', 'filter_mutate', 'get_svn_revision', 'InstallList',
-           'make_install_path', 'wwwuid']
+           'wwwuid']
 
 # Determine which Python version (2.4 or 2.5, for example) we are running,
 # and use that as the filename to the Python directory.
@@ -311,17 +311,6 @@
             _, pylist[i] = pylist[i].split(os.sep, 1)
     return pylist
 
-def make_install_path(rootdir, path):
-    '''Combine an installation root directory and final install path.
-
-    Normalises path, and joins it to the end of rootdir, removing the leading
-    / to make it relative if required.
-    '''
-    normpath = os.path.normpath(path)
-    if normpath.startswith(os.sep):
-        normpath = normpath[1:]
-    return os.path.join(rootdir, normpath)
-
 class InstallList(object):
     # We build two separate lists, by walking www and console
     list_www = property(lambda self: build_list_py_files('www'))
@@ -342,7 +331,6 @@
         "services/usrmgt-server",
         "services/diffservice",
         "services/svnlogservice",
-        "services/usrmgt-server", # XXX: Should be in bin/
     ]
 
     list_user_binaries = [
@@ -354,5 +342,4 @@
         "bin/ivle-mountallusers",
         "bin/ivle-remakeuser",
         "bin/ivle-showenrolment",
-        "bin/ivle-buildjail",
     ]

=== modified file 'www/apps/help/__init__.py'
--- www/apps/help/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/help/__init__.py	2009-02-02 04:00:25 +0000
@@ -28,6 +28,9 @@
 from ivle import util
 import ivle.conf
 
+import genshi
+import genshi.template
+
 # TODO: Nicer 404 errors
 
 def handle(req):
@@ -50,28 +53,28 @@
 
 def show_help_menu(req):
     """Show the help menu."""
-
     # Set request attributes
     req.content_type = "text/html"
     req.write_html_head_foot = True
 
-    # Start writing data
-    req.write('<div id="ivle_padding">\n')
-    req.write("<h1>Help</h1>\n")
+    ctx = genshi.template.Context()
 
     # Write a list of links to all apps with help modules
-    req.write("<ul>\n")
+    ctx['apps'] = []
     # Tab apps, in order of tabs
     for appurl in ivle.conf.apps.apps_in_tabs:
         app = ivle.conf.apps.app_url[appurl]
         if app.hashelp:
-            req.write('  <li><a href="%s">%s</a></li>\n'
-                % (os.path.join(util.make_path("help"), appurl), app.name))
+            new_app = {}
+            new_app['appurl'] = os.path.join(util.make_path("help"), appurl)
+            new_app['name'] = app.name
+            ctx['apps'].append(new_app)
     # Terms of Service
-    req.write('  <li><a href="%s">Terms of Service</a></li>\n'
-        % (util.make_path("tos")))
-    req.write("</ul>\n")
-    req.write('</div>\n')
+    ctx['tos'] = util.make_path("tos")
+
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("apps/help/template.html"))
+    req.write(tmpl.generate(ctx).render('html')) #'xhtml', doctype='xhtml'))
 
 def show_help_app(req, app):
     """Show help for an application."""

=== added file 'www/apps/help/template.html'
--- www/apps/help/template.html	1970-01-01 00:00:00 +0000
+++ www/apps/help/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,17 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+
+  <body py:strip="">
+  
+    <div id="ivle_padding">
+      <h1>Help</h1>
+      <ul>
+        <py:for each="app in apps">
+          <li><a href="${app['appurl']}">${app['name']}</a></li>
+        </py:for>
+        <li><a href="${tos}">Terms of Service</a></li>
+      </ul>
+    </div>
+  </body>
+</html>

=== modified file 'www/apps/logout/__init__.py'
--- www/apps/logout/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/logout/__init__.py	2009-02-02 04:00:25 +0000
@@ -19,9 +19,11 @@
 # Author: Nick Chadwick
 # Date: 13/01/2009
 
-import cgi
 from ivle import util
 
+import genshi
+import genshi.template
+
 # url path for this app
 THIS_APP = "logout"
 
@@ -29,10 +31,10 @@
     if req.method == "POST":
         req.logout()
     else:
+        req.content_type = "text/html"
         req.write_html_head_foot = True
-        req.content_type = "text/html"
-        req.write('<div id="ivle_padding">\n'
-                  '<h3>Are you sure you want to logout?</h3><p>'
-                  '<form action="%s" method="POST">\n'
-                  '    <input type="submit" value="Logout" />\n'
-                  '</form>\n</div>\n' % (cgi.escape(util.make_path('logout'))))
+        ctx = genshi.template.Context()
+        ctx['path'] =  util.make_path('logout')
+        loader = genshi.template.TemplateLoader(".", auto_reload=True)
+        tmpl = loader.load(util.make_local_path("apps/logout/template.html"))
+        req.write(tmpl.generate(ctx).render('html')) #'xhtml', doctype='xhtml'))

=== added file 'www/apps/logout/template.html'
--- www/apps/logout/template.html	1970-01-01 00:00:00 +0000
+++ www/apps/logout/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,15 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+
+  <body py:strip="">
+    <div id="ivle_padding">
+      <h3>Are you sure you want to logout?</h3>
+      <p>
+        <form action="${path}" method="POST">
+          <input type="submit" value="Logout" />
+        </form>
+      </p>
+    </div>
+  </body>
+</html>

=== modified file 'www/apps/settings/__init__.py'
--- www/apps/settings/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/settings/__init__.py	2009-02-02 04:00:25 +0000
@@ -24,6 +24,9 @@
 
 from ivle import util
 
+import genshi
+import genshi.template
+
 def handle(req):
     """Handler for the Settings application."""
 
@@ -44,30 +47,6 @@
     ]
     req.write_html_head_foot = True     # Have dispatch print head and foot
 
-    # Start writing data
-    req.write("""<div id="ivle_padding">
-
-  <h2>User Profile</h2>
-  <p><span id="login"></span></p>
-  <p id="role"></p>
-  <h3>Change settings</h3>
-  <form action="">
-    <table>
-      <tr><td>Display name:</td><td><input type="text"
-            name="nick" id="nick" size="40" /></td></tr>
-      <tr><td>Email address:</td><td><input type="text"
-            name="email" id="email" size="40" /></td></tr>
-    </table>
-    <div id="changepassword"></div>
-    <p id="result"></p>
-    <p>
-      <input value="Save" onclick="return save_settings()" id="save"
-          type="submit" />
-      <input value="Revert" onclick="return revert_settings()" id="revert"
-          type="reset" />
-    </p>
-  </form>
-
-  <div id="notices"></div>
-</div>
-""")
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("apps/settings/template.html"))
+    req.write(tmpl.generate().render('html')) #'xhtml', doctype='xhtml'))

=== added file 'www/apps/settings/template.html'
--- www/apps/settings/template.html	1970-01-01 00:00:00 +0000
+++ www/apps/settings/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,43 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+
+  <body py:strip="">
+    <div id="ivle_padding">
+
+      <h2>User Profile</h2>
+      <p>
+        <span id="login"></span>
+      </p>
+      <p id="role"></p>
+      <h3>Change settings</h3>
+      <form action="">
+        <table>
+          <tr>
+            <td>Display name:</td>
+            <td>
+              <input type="text"
+                  name="nick" id="nick" size="40" />
+            </td>
+          </tr>
+          <tr>
+            <td>Email address:</td>
+            <td>
+              <input type="text"
+                  name="email" id="email" size="40" />
+            </td>
+          </tr>
+        </table>
+        <div id="changepassword"></div>
+        <p id="result"></p>
+        <p>
+          <input value="Save" onclick="return save_settings()" id="save"
+                type="submit" />
+          <input value="Revert" onclick="return revert_settings()" id="revert"
+                type="reset" />
+        </p>
+      </form>
+      <div id="notices"></div>
+    </div>
+  </body>
+</html>

=== removed directory 'www/apps/subjects'
=== modified file 'www/apps/userservice/__init__.py'
--- www/apps/userservice/__init__.py	2009-02-02 03:29:40 +0000
+++ www/apps/userservice/__init__.py	2009-02-04 12:51:17 +0000
@@ -142,7 +142,6 @@
 import datetime
 
 import cjson
-import pg
 
 import ivle.database
 from ivle import (util, chat, caps)
@@ -361,21 +360,15 @@
 
     user = ivle.database.User.get_by_login(req.store, login)
 
-    # Make a dict of fields to update
     oldpassword = fields.getfirst('oldpass')
-    for f in fieldlist:
-        val = fields.getfirst(f)
-        if val is not None:
-            # Note: May be rolled back if auth check below fails
-            setattr(user, f, val.value.decode('utf-8'))
-        else:
-            pass
+    if oldpassword is not None: # It was specified.
+        oldpassword = oldpassword.value
 
     # If the user is trying to set a new password, check that they have
     # entered old password and it authenticates.
     if fields.getfirst('password') is not None:
         try:
-            authenticate.authenticate(login, oldpassword)
+            authenticate.authenticate(req.store, login, oldpassword)
         except AuthError:
             req.headers_out['X-IVLE-Action-Error'] = \
                 urllib.quote("Old password incorrect.")
@@ -384,6 +377,15 @@
             req.store.rollback()
             return
 
+    # Make a dict of fields to update
+    for f in fieldlist:
+        val = fields.getfirst(f)
+        if val is not None:
+            # Note: May be rolled back if auth check below fails
+            setattr(user, f, val.value.decode('utf-8'))
+        else:
+            pass
+
     req.store.commit()
 
     req.content_type = "text/plain"

=== modified file 'www/plugins/console/__init__.py'
--- www/plugins/console/__init__.py	2009-01-13 01:36:15 +0000
+++ www/plugins/console/__init__.py	2009-02-02 04:00:25 +0000
@@ -30,6 +30,10 @@
 
 from ivle import util
 
+import genshi
+import genshi.core
+import genshi.template
+
 def insert_scripts_styles(scripts, styles, scripts_init):
     """Given 2 lists of strings: scripts and styles. These lists are lists of
     pathnames, as expected on the "scripts" and "styles" attributes of a
@@ -53,45 +57,14 @@
     windowpane: If True, starts the console in "window pane" mode, where it
     will float over the page and have a "minimize" button.
     """
-    req.write("""<div id="console_body">
-  <div id="console_heading">Python Console
-""")
-    if windowpane:
-        req.write("""<span class="console_button minimize">
-      <a onclick="console_minimize()"
-        title="Minimize the Python console">
-        <img src="%s" /></a>
-    </span>
-""" % cgi.escape(util.make_path("media/images/interface/minimize.png")))
-    req.write("""</div>
-  <div id="console_body2">
-  <pre id="console_output" class='outputMsg'></pre>
-  <div id="console_input">
-    <div id="console_inputArea">
-    </div>
-    <label id="console_prompt">&gt;&gt;&gt; </label>
-    <input id="console_inputText"
-      type="text" size="80" onkeypress="return catch_input(event.keyCode)" />
-    <input type='button' value='Interrupt' onclick='set_interrupt();'/>
-    <input type='button' value='Clear Output' onclick='return clear_output();'/>
-    <input type='button' value='Reset' onclick='console_reset();'/>
-""")
-    if windowpane:
-        req.write("""<span class="console_button maximize">
-      <a onclick="console_maximize()"
-        title="Open up the Python console">
-        <img src="%s" /></a>
-    </span>
-""" % cgi.escape(util.make_path("media/images/interface/maximize.png")))
-    req.write("""</div>
-</div></div>
-""")
-    if windowpane:
-        req.write("""
-<!-- Console filler, provides extra vertical space to stop the console
-     covering over the bottom content -->
-<div id="console_filler"></div>
-""")
+    ctx = genshi.template.Context()
+    ctx['windowpane'] = windowpane
+    ctx['minimize_path'] = util.make_path("media/images/interface/minimize.png") 
+    ctx['maximize_path'] = util.make_path("media/images/interface/maximize.png")
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("plugins/console/template.html"))
+    #TODO: Make the dispatch render this
+    req.write(tmpl.generate(ctx).render('html'))
 
 def _append_if_absent(list, *values):
     """Appends an arbitray number of values to a list, but omits values that

=== added file 'www/plugins/console/template.html'
--- www/plugins/console/template.html	1970-01-01 00:00:00 +0000
+++ www/plugins/console/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,40 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+      
+  <body py:strip="">
+    <div id="console_body">
+      <div id="console_heading">Python Console<span py:if="windowpane" 
+          class="console_button minimize">
+          <a onclick="console_minimize()"
+              title="Minimize the Python console">
+            <img src="${minimize_path}" />
+          </a>
+        </span>
+      </div>
+      <div id="console_body2">
+        <pre id="console_output" class='outputMsg'></pre>
+        <div id="console_input">
+          <div id="console_inputArea">
+          </div>
+          <label id="console_prompt">&gt;&gt;&gt; </label>
+          <input id="console_inputText" type="text" size="80" 
+              onkeypress="return catch_input(event.keyCode)" />
+          <input type='button' value='Interrupt' onclick='set_interrupt();'/>
+          <input type='button' value='Clear Output' onclick='return clear_output();'/>
+          <input type='button' value='Reset' onclick='console_reset();'/>
+          <span py:if="windowpane" class="console_button maximize">
+            <a onclick="console_maximize()"
+                title="Open up the Python console">
+              <img src="${maximize_path}" />
+            </a>
+          </span>
+        </div>
+      </div>
+    </div>
+    <div py:if="windowpane" id="console_filler">
+    <!-- Console filler, provides extra vertical space to stop the console
+                                            covering over the bottom content -->
+    </div>
+  </body>
+</html>

